char program[] = "cudaWagstaff v0.03"; //do not change

/* Adapted from CUDALucas */

/* CUDALucas.c
   Shoichiro Yamada Oct. 2010

   This is an adaptation of Richard Crandall lucdwt.c, John Sweeney MacLucasUNIX.c,
   and Guillermo Ballester Valor MacLucasFFTW.c code.
   Improvement From Prime95.

   It also contains mfaktc code by Oliver Weihe and Eric Christenson
   adapted for CUDALucas use. Such code is under the GPL, and is noted as such.
*/

/* Include Files */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <time.h>
#include <signal.h>
#ifndef _MSC_VER
#include <sys/types.h>
#include <sys/stat.h>
#else
#include <direct.h>
#endif
#include <cuda.h>

// Check windows
#if _WIN32 || _WIN64
#if _WIN64
#define WIN_ENVIRONMENT64
#include <nvml.h>
#else
#define WIN_ENVIRONMENT32
#endif
#endif

// Check GCC
#if __GNUC__
#if __x86_64__ || __ppc64__
#define LINUX_ENVIRONMENT64
#else
#define LINUX_ENVIRONMENT32
#endif
#endif

#include <cuda_runtime.h>
#include <cufft.h>
#include <cuComplex.h>
#include "cuda_safecalls.h"
#include "parse.h"


/* In order to have the gettimeofday() function, you need these includes on Linux:
#include <sys/time.h>
#include <unistd.h>
On Windows, you need
#include <winsock2.h> and a definition for
int gettimeofday (struct timeval *tv, struct timezone *) {}
Both platforms are taken care of in parse.h and parse.c. */
/* http://www.kurims.kyoto-u.ac.jp/~ooura/fft.html
   base code from Takuya OOURA.  */

/************************ definitions ************************************/
#ifdef _MSC_VER
#define strncasecmp strnicmp
#endif

/**************************************************************
***************************************************************
*                                                             *
*                       Global Variables                      *
*                                                             *
***************************************************************
**************************************************************/
cufftDoubleComplex *d_x;          //device FFT data
cufftDoubleComplex *h_x_debug;

cufftDoubleComplex *d_negacyclic_weight; // device FFT weight data. Will be used for both weighting and unweighting
double2 *d_irrational_weight;     // forward irrational weight
double2 *d_irrational_weight_inverse;    // inverse of forward irrational weight
int2    *d_xint;                  //integer copy of gpu data for transfer to cpu
int2    *h_xint_debug;
//int2    *h_xint_simulate;
char2   *d_base_bit;              // how many bits of the number per limb of fft

int2  *d_block_carry;        // inter-block carry data
int2  *h_bc_debug;
//longlong2  *h_bc_simulate;
__device__ float g_err;           //current maximum error

char2   *h_base_bit;              // host version of d_base_bit

int drv_ver;            //CUDA Driver Version
int rt_ver;              //CUDA Run-Time Version

cufftHandle    g_plan;
cudaDeviceProp g_dev;             //structure holding device property information

int g_ffts[256];                  //array of ffts lengths
int g_fft_count = 0;              //number of ffts lengths in g_ffts
int g_fftlen;                     //index of the fft currently in use
int g_lbi = 0;                    //index of the smallest feasible fft length for the current exponent
int g_ubi = 256;                  //index of the largest feasible fft length for the current exponent
int g_cpi;                        //checkpoint interval
int g_ri;                         //report interval
int g_ei;                         //error check interval
int g_pf;                         //polite flag
int g_po;                         //polite value
int g_qu;                         //quitting flag
int g_dn;                         //device number
int g_df;                         //show device info flag
int g_ki;                         //keyboard input flag


char g_folder[192];               //folder where savefiles will be kept
char g_input_file[192];           //input file name default worktodo.txt
char g_RESULTSFILE[192];          //output file name default results.txt
char g_INIFILE[192] = "cudaWagstaff.ini"; //initialization file name
char g_output[50][32];
char g_output_string[192];
char g_output_header[192];
int  g_output_code[50];
int  g_output_interval;

typedef struct gec_data
{
  int L1;
  int L2;
  int last_check_iteration;
  int last_cumulative_iteration;
  int accumulate_count;
  int rollback_count_total;
  int rollback_count_this_pass;
  cufftDoubleComplex *d_x_scratch;
  cufftDoubleComplex *d_base; // u(0)
  cufftDoubleComplex *d_base_scratch; // temp area for check calculation
  int2 *h_base_int;
  cufftDoubleComplex *d_cumulative; //d(t)
  int2 *d_cumulative_int; // int version of d(t)
  cufftDoubleComplex *d_scratch_cumulative; // temp area for check calculation
  int2 *h_cumulative_int; // computed using accumulation
  int2 *h_cumulative_int_scratch; // computed using verification
} GEC_DATA;

GEC_DATA gec;

/**************************************************************
***************************************************************
*                                                             *
*             Kernels and other device functions              *
*                                                             *
***************************************************************
**************************************************************/


/**************************************************************
*                                                             *
*                       Device functions                      *
*                                                             *
**************************************************************/

// inline ptx rounding function
# define RINT(x)  __rintd(x)

__device__ static double __rintd (double z)
{
  double y;
  asm ("cvt.rni.f64.f64 %0, %1;": "=d" (y):"d" (z));
  return (y);
}

/**************************************************************
*                                                             *
*                           Kernels                           *
*                                                             *
**************************************************************/

// The pointwise squaring of the fft'ed data.
__global__ void square (int n, cufftDoubleComplex * __restrict__ a)
{
  int j = blockIdx.x * blockDim.x + threadIdx.x;

  cufftDoubleComplex val = a[j];
  cufftDoubleComplex tmp;

  tmp.x = (val.x + val.y)*(val.x - val.y);
  tmp.y = 2.0*val.x*val.y;

  a[j] = tmp;
}

// The pointwise multiplication of the fft'ed data.
// a = a*b
__global__ void multiply (const int n, cufftDoubleComplex * __restrict__ a, const cufftDoubleComplex * __restrict__ b, const int bkp_flag, cufftDoubleComplex * __restrict__ c)
{
  int j = blockIdx.x * blockDim.x + threadIdx.x;

  cufftDoubleComplex val = a[j];
  cufftDoubleComplex val2 = b[j];
  if(bkp_flag) c[j]=val;
  cufftDoubleComplex tmp;


  tmp.x = val.x*val2.x - val.y*val2.y;
  tmp.y = val.x*val2.y + val.y*val2.x;

  a[j] = tmp;
}


__global__ void weight(const int n, const int2 * __restrict__ aint, cufftDoubleComplex * __restrict__ a, const cufftDoubleComplex * __restrict__ neg_wt, const double2 * __restrict__ ib_wt)
{
  int j = blockIdx.x * blockDim.x + threadIdx.x;

  int2 val = aint[j]; // read the balanced int representation
  double2 ib = ib_wt[j];
  cufftDoubleComplex nwt = neg_wt[j];
  cufftDoubleComplex tmp;

  // first, apply the irrational base weights
  tmp.x = val.x * ib.x;
  tmp.y = val.y * ib.y;
  // then apply the negacyclic weight and save
  a[j] = make_double2( tmp.x*nwt.x - tmp.y*nwt.y, tmp.y*nwt.x + tmp.x*nwt.y );
}

/*
__global__ void unweight_and_partial_carry(int n, cufftDoubleComplex * __restrict__ a, const cufftDoubleComplex * __restrict__ neg_wt, const double2 * __restrict__ ib_inv_wt, int2 * __restrict__ aint, const char2 * __restrict__ base_bit, longlong2 * __restrict__ block_carry)
{
  int j = blockIdx.x * blockDim.x + threadIdx.x;
  cufftDoubleComplex val = a[j];
  double2 ib = ib_inv_wt[j];
  cufftDoubleComplex nwt = neg_wt[j];
  longlong2 tmp;

  // undo negacyclic weight. use -nwt.y to get 1/nwt
  // then undo ib weight
  // then round
  // then save, in preparation for carry propagation
  double2 u = make_double2( (val.x*nwt.x + val.y*nwt.y)*ib.x, (val.y*nwt.x - val.x*nwt.y)*ib.y );
  double2 roundu = make_double2( RINT( u.x ), RINT( u.y ) );
  a[j] = roundu;
  tmp.x = (long long) roundu.x;
  tmp.y = (long long) roundu.y;

  float err_max = max( fabs ((float) (u.x - roundu.x)), fabs ((float) (u.y - roundu.y)) );
  if(err_max > 0.0625f) atomicMax ((int*) &g_err, __float_as_int (err_max));
  

  char2 bits = base_bit[j];
  int2 base = make_int2( 1 << bits.x, 1 << bits.y );
  int2 mask = make_int2( base.x-1, base.y-1 );
  int2 cmp = make_int2( base.x >> 1, base.y >> 1 );

  __shared__ longlong2 carry[128];

  longlong2 carry_out_block; // final thread in block maintains block carry out and writes to global memory

  carry[threadIdx.x] = carry_out_block = make_longlong2( tmp.x >> bits.x, tmp.y >> bits.y ); // carry is no larger than 53 - 16 = 37 bits. probably too conservative.
  __threadfence_block();
  __syncthreads();

  tmp.x &= mask.x;
  tmp.y &= mask.y;


  if( threadIdx.x > 0 ) // lowest thread in the block has nothing further to do
  {
    tmp.x += carry[threadIdx.x-1].x; // sum is no larger than 37+1 bits. probably too conservative.
    tmp.y += carry[threadIdx.x-1].y;
    __threadfence_block();
  }
  __syncthreads();
  if( threadIdx.x == 127) // final thread in block
  {
    carry_out_block.x += (tmp.x >> bits.x);
    carry_out_block.y += (tmp.y >> bits.y);
  }

   // carry is no larger than 38 -16 = 22. safe to start using int instead of long long

  __shared__ int2 icarry[128]; // is compiler intelligent enough to reuse shared memory from longlong2 carry[128] ?

  icarry[threadIdx.x] = make_int2( (int) (tmp.x >> bits.x), (int) (tmp.y >> bits.y) );
  __threadfence_block();
  __syncthreads();

  int2 itmp;

  itmp.x = (int) tmp.x & mask.x;
  itmp.y = (int) tmp.y & mask.y;

  if( threadIdx.x > 0 )
  {
    itmp.x += icarry[threadIdx.x-1].x; // 22+1
    itmp.y += icarry[threadIdx.x-1].y;
    __threadfence_block();
  } 
  __syncthreads();
  int2 icarry_out = make_int2( itmp.x >> bits.x, itmp.y >> bits.y ); // 23-16=7

  itmp.x &= mask.x;
  itmp.y &= mask.y;

  if(itmp.x >= cmp.x) { icarry_out.x++; itmp.x -= base.x; } // balance the digits. crudely.
  if(itmp.y >= cmp.y) { icarry_out.y++; itmp.y -= base.y; }

  if( threadIdx.x == 127)
  {
    carry_out_block.x += (long long) icarry_out.x;
    carry_out_block.y += (long long) icarry_out.y;
  }

  icarry[threadIdx.x] = icarry_out;
  __threadfence_block();
  __syncthreads();

  if( threadIdx.x > 0 )
  {
    itmp.x += icarry[threadIdx.x-1].x; // carry is small. don't bother with further balancing.
    itmp.y += icarry[threadIdx.x-1].y;
    __threadfence_block();
  }
  __syncthreads();
  if( threadIdx.x == 127) // write out block carry to global
  {
    block_carry[blockIdx.x] = carry_out_block;
    __threadfence_block();
  }
  __syncthreads();

  aint[j] = itmp; // save to global int array. next step, we'll convert back to double.
}
*/

/*
__global__ void unweight_and_partial_carry_unsafe(int n, cufftDoubleComplex * __restrict__ a, const cufftDoubleComplex * __restrict__ neg_wt, const double2 * __restrict__ ib_inv_wt, int2 * __restrict__ aint, const char2 * __restrict__ base_bit, int2 * __restrict__ block_carry)
{
  int j = blockIdx.x * blockDim.x + threadIdx.x;
  cufftDoubleComplex val = a[j];
  double2 ib = ib_inv_wt[j];
  cufftDoubleComplex nwt = neg_wt[j];
  longlong2 tmp;


  // undo negacyclic weight. use -nwt.y to get 1/nwt
  // then undo ib weight
  // then round
  // then save, in preparation for carry propagation
  double2 u = make_double2( (val.x*nwt.x + val.y*nwt.y)*ib.x, (val.y*nwt.x - val.x*nwt.y)*ib.y );
  double2 roundu = make_double2( RINT( u.x ), RINT( u.y ) );
  a[j] = roundu;
  tmp.x = (long long) roundu.x;
  tmp.y = (long long) roundu.y;

  float err_max = max( fabs ((float) (u.x - roundu.x)), fabs ((float) (u.y - roundu.y)) );
  if(err_max > 0.0625f) atomicMax ((int*) &g_err, __float_as_int (err_max));
  

  char2 bits = base_bit[j];
  int2 base = make_int2( 1 << bits.x, 1 << bits.y );
  int2 mask = make_int2( base.x-1, base.y-1 );
  int2 cmp = make_int2( base.x >> 1, base.y >> 1 );

  __shared__ int2 icarry[128];

  int2 carry_out_block; // final thread in block maintains block carry out and writes to global memory

  icarry[threadIdx.x] = carry_out_block = make_int2( (int) (tmp.x >> bits.x), (int) (tmp.y >> bits.y) );
  __threadfence_block();
  __syncthreads();

  int2 itmp;
  itmp.x = (int) tmp.x & mask.x;
  itmp.y = (int) tmp.y & mask.y;


  if( threadIdx.x > 0 ) // lowest thread in the block has nothing further to do
  {
    itmp.x += icarry[threadIdx.x-1].x; // sum is no larger than 37+1 bits. probably too conservative.
    itmp.y += icarry[threadIdx.x-1].y;
    __threadfence_block();
  }
  __syncthreads();

  int2 c1 = make_int2( itmp.x >> bits.x, itmp.y >> bits.y );
  itmp.x &= mask.x;
  itmp.y &= mask.y;

  if( threadIdx.x == 127) // final thread in block
  {
    carry_out_block.x += c1.x;
    carry_out_block.y += c1.y;
  }

  icarry[threadIdx.x] = c1;
  __threadfence_block();
  __syncthreads();

  if( threadIdx.x > 0 )
  {
    itmp.x += icarry[threadIdx.x-1].x; // 22+1
    itmp.y += icarry[threadIdx.x-1].y;
    __threadfence_block();
  } 
  __syncthreads();
  int2 icarry_out = make_int2( itmp.x >> bits.x, itmp.y >> bits.y ); // 23-16=7

  itmp.x &= mask.x;
  itmp.y &= mask.y;

  if(itmp.x >= cmp.x) { icarry_out.x++; itmp.x -= base.x; } // balance the digits. crudely.
  if(itmp.y >= cmp.y) { icarry_out.y++; itmp.y -= base.y; }

  if( threadIdx.x == 127)
  {
    carry_out_block.x += icarry_out.x;
    carry_out_block.y += icarry_out.y;
  }

  icarry[threadIdx.x] = icarry_out;
  __threadfence_block();
  __syncthreads();

  if( threadIdx.x > 0 )
  {
    itmp.x += icarry[threadIdx.x-1].x; // carry is small. don't bother with further balancing.
    itmp.y += icarry[threadIdx.x-1].y;
    __threadfence_block();
  }
  __syncthreads();
  if( threadIdx.x == 127) // write out block carry to global
  {
    block_carry[blockIdx.x] = carry_out_block;
    __threadfence_block();
  }
  __syncthreads();

  aint[j] = itmp; // save to global int array. next step, we'll convert back to double.
}
*/

__global__ void unweight_and_partial_carry_unsafe_opt(int n, const cufftDoubleComplex * __restrict__ a, const cufftDoubleComplex * __restrict__ neg_wt, const double2 * __restrict__ ib_inv_wt, int2 * __restrict__ aint, const char2 * __restrict__ base_bit, int2 * __restrict__ block_carry)
{
  const int j = blockIdx.x * blockDim.x + threadIdx.x;
  const cufftDoubleComplex val = a[j];
  const double2 ib = ib_inv_wt[j];
  const cufftDoubleComplex nwt = neg_wt[j];
  longlong2 tmp;
 
  // undo negacyclic weight. use -nwt.y to get 1/nwt
  // then undo ib weight
  // then round
  // then save, in preparation for carry propagation
  const double2 u = make_double2( (val.x*nwt.x + val.y*nwt.y)*ib.x, (val.y*nwt.x - val.x*nwt.y)*ib.y );
  //const double2 roundu = make_double2( RINT( u.x ), RINT( u.y ) );
  //a[j] = roundu;
  tmp.x = (long long) RINT( u.x );
  tmp.y = (long long) RINT( u.y );

/*
  float err_max = max( fabs ((float) (u.x - roundu.x)), fabs ((float) (u.y - roundu.y)) );
  if(err_max > 0.0625f) atomicMax ((int*) &g_err, __float_as_int (err_max));
*/

  const char2 bits = base_bit[j];
  const int2 base = make_int2( 1 << bits.x, 1 << bits.y );
  const int2 mask = make_int2( base.x-1, base.y-1 );
  const int2 cmp = make_int2( base.x >> 1, base.y >> 1 );

  int2 carry_out_block; // final thread in block maintains block carry out and writes to global memory

  __shared__ int2 icarry[128+1];

  if(threadIdx.x == 0) icarry[0] = make_int2(0,0);
  icarry[threadIdx.x+1] = carry_out_block = make_int2( (int) (tmp.x >> bits.x), (int) (tmp.y >> bits.y) ); // assume 31 bits max
  __syncthreads();

  int2 itmp;
  itmp.x = (int) tmp.x & mask.x;
  itmp.y = (int) tmp.y & mask.y;


  itmp.x += icarry[threadIdx.x].x;
  itmp.y += icarry[threadIdx.x].y;
  __syncthreads();

  int2 icarry_out = make_int2( itmp.x >> bits.x, itmp.y >> bits.y ); // 23-16=7

  itmp.x &= mask.x;
  itmp.y &= mask.y;

  if(itmp.x >= cmp.x) { icarry_out.x++; itmp.x -= base.x; } // balance the digits. crudely.
  if(itmp.y >= cmp.y) { icarry_out.y++; itmp.y -= base.y; }

  carry_out_block.x += icarry_out.x;
  carry_out_block.y += icarry_out.y;

  icarry[threadIdx.x+1] = icarry_out;
  __syncthreads();

  itmp.x += icarry[threadIdx.x].x; // carry is small. don't bother with further balancing.
  itmp.y += icarry[threadIdx.x].y;

  aint[j] = itmp; // save to global int array. next step, we'll convert back to double.

  if( threadIdx.x == 127) // write out block carry to global
  {
    block_carry[blockIdx.x] = carry_out_block;
  }

}
__global__ void finish_carry_and_reweight(int n, cufftDoubleComplex * __restrict__ a, const cufftDoubleComplex * __restrict__ neg_wt, const double2 * __restrict__ ib_wt, int2 * __restrict__ aint, const char2 * __restrict__ base_bit, const int2 * __restrict__ block_carry)
{
  const int j = blockIdx.x * blockDim.x + threadIdx.x;

  __shared__ int2 a_updated[128];

  a_updated[threadIdx.x] = aint[j];
  //__syncthreads();
  __syncwarp();

  if(threadIdx.x == 0) // only thread 0 of each block will spread the block carry to the next 2 locations
  {
    int2 block_carry_in;
    if( blockIdx.x == 0) // first block uses last blocks carry out as its in.
    {
      block_carry_in = block_carry[ gridDim.x-1 ];
      // last carry's x component is carry in for 0th blocks y component. and y component is negated for 0th component's x carry in
      block_carry_in = make_int2( -block_carry_in.y, block_carry_in.x );
    }
    else
    {
      block_carry_in = block_carry[ blockIdx.x-1 ];
    }
    
    int2 a0, a1, a2;
    char2 bits0, bits1;
    int2 mask0, mask1;
    int2 base0, base1;
    int2 cmp0, cmp1;

    bits0 = base_bit[j+0]; base0 = make_int2(1 << bits0.x, 1 << bits0.y); mask0 = make_int2( base0.x-1, base0.y-1 ); cmp0 = make_int2( base0.x >> 1, base0.y >> 1 );
    bits1 = base_bit[j+1]; base1 = make_int2(1 << bits1.x, 1 << bits1.y); mask1 = make_int2( base1.x-1, base1.y-1 ); cmp1 = make_int2( base1.x >> 1, base1.y >> 1 );

    a0 = a_updated[0];
    a1 = a_updated[1];
    a2 = a_updated[2];

    int2 tmp;
    
    tmp.x = a0.x + block_carry_in.x; tmp.y = a0.y + block_carry_in.y; // 38
    a1.x += tmp.x >> bits0.x; a1.y += tmp.y >> bits0.y; // 22
    a0.x = tmp.x & mask0.x; a0.y = tmp.y & mask0.y;
    if (a0.x >= cmp0.x) { a1.x++; a0.x -= base0.x; }
    if (a0.y >= cmp0.y) { a1.y++; a0.y -= base0.y; }
    a2.x += (a1.x >> bits1.x); a2.y += (a1.y >> bits1.y); // max 6 bits of carry. don't bother adjusting
    a1.x &= mask1.x; a1.y &= mask1.y;
    if (a1.x >= cmp1.x) { a2.x++; a1.x -= base1.x; }
    if (a1.y >= cmp1.y) { a2.y++; a1.y -= base1.y; }

    
    aint[j+0] = a_updated[0] = a0;
    aint[j+1] = a_updated[1] = a1;
    aint[j+2] = a_updated[2] = a2;
    //__threadfence_block();
  }
  //__syncthreads();
  __syncwarp();

  // reweight and save back to fft

  const int2 val = a_updated[threadIdx.x];

  cufftDoubleComplex dval;
  double2 ib = ib_wt[j];
  cufftDoubleComplex nwt = neg_wt[j];
  cufftDoubleComplex tmp;

  // first, apply the irrational base weights
  dval.x = val.x * ib.x;
  dval.y = val.y * ib.y;
  // then apply the negacyclic weight.
  tmp.x = dval.x*nwt.x - dval.y*nwt.y;
  tmp.y = dval.y*nwt.x + dval.x*nwt.y;

  a[j] = tmp;
}


/**************************************************************
***************************************************************
*                                                             *
*                        Host functions                       *
*                                                             *
***************************************************************
**************************************************************/
/* Take rounded output a[] and simulate carry propagation */
/*
void simulate_unweight_and_partial_carry(const int n, cufftDoubleComplex * a, int2 * aint, const char2 * base_bit, longlong2 * block_carry)
{
  longlong2 carry[n/2];
  int2 icarry[n/2];

  // First pass
  for(int j=0; j < n/2; j++)
  {
    int blockId = j / 128;
    int threadId = j % 128;
    longlong2 tmp;

    char2 bits = base_bit[j];
    int2 base = make_int2( 1 << bits.x, 1 << bits.y );
    int2 mask = make_int2( base.x-1, base.y-1 );
    int2 cmp = make_int2( base.x >> 1, base.y >> 1 );

    tmp.x = (long long) a[j].x;
    tmp.y = (long long) a[j].y;

    carry[j] = make_longlong2( tmp.x >> bits.x, tmp.y >> bits.y );
    if (threadId == 127)
      block_carry[blockId] = carry[j];

    tmp.x &= mask.x;
    tmp.y &= mask.y;


    if( threadId > 0 )
    {
      tmp.x += carry[j-1].x;
      tmp.y += carry[j-1].y;
    }
    if( threadId == 127)
    {
      block_carry[blockId].x += (tmp.x >> bits.x);
      block_carry[blockId].y += (tmp.y >> bits.y);
    }

    icarry[j] = make_int2( (int) (tmp.x >> bits.x), (int) (tmp.y >> bits.y) );

    aint[j].x = (int) tmp.x & mask.x;
    aint[j].y = (int) tmp.y & mask.y;

    if( threadId > 0 )
    {
      aint[j].x += icarry[j-1].x;
      aint[j].y += icarry[j-1].y;
    } 
  }

  // Second pass
  for(int j=0; j < n/2; j++)
  {
    int blockId = j / 128;
    int threadId = j % 128;

    char2 bits = base_bit[j];
    int2 base = make_int2( 1 << bits.x, 1 << bits.y );
    int2 mask = make_int2( base.x-1, base.y-1 );
    int2 cmp = make_int2( base.x >> 1, base.y >> 1 );

    int2 icarry_out = make_int2( aint[j].x >> bits.x, aint[j].y >> bits.y );

    aint[j].x &= mask.x;
    aint[j].y &= mask.y;

    if(aint[j].x >= cmp.x) { icarry_out.x++; aint[j].x -= base.x; } // balance the digits. crudely.
    if(aint[j].y >= cmp.y) { icarry_out.y++; aint[j].y -= base.y; }

    if( threadId == 127)
    {
      block_carry[blockId].x += (long long) icarry_out.x;
      block_carry[blockId].y += (long long) icarry_out.y;
    }

    icarry[j] = icarry_out;

    if( threadId > 0 )
    {
      aint[j].x += icarry[j-1].x; // carry is small. don't bother with further balancing.
      aint[j].y += icarry[j-1].y;
    }
  }
}
*/

/**************************************************************
*                                                             *
*                        Initialization                       *
*                                                             *
**************************************************************/

void init_device (int device_number, int show_prop)
{
  int device_count = 0;
  cudaGetDeviceCount (&device_count);
  if (device_number >= device_count)
  {
  printf ("device_number %d >=  device_count %d ... exiting\n", device_number, device_count);
  printf ("(If you didn't select a device this is probably a driver problem)\n\n");
  exit (2);
  }
  cudaSetDevice (device_number);
  cutilSafeCall(cudaSetDeviceFlags (cudaDeviceBlockingSync));
  cudaGetDeviceProperties (&g_dev, device_number);
  cudaRuntimeGetVersion(&rt_ver);
  cudaDriverGetVersion(&drv_ver);
  
// From Iain
if (g_dev.major == 1 && g_dev.minor < 3)
  {
  printf("A GPU with compute capability >= 1.3 is required for double precision arithmetic\n\n");
    printf("See http://www.mersenne.ca/cudalucas.php for a list of cards\n\n");
  exit (2);
  }

#if CUDART_VERSION >= 7000  //CUDA >=7 Does not support compute_1.x
  if (g_dev.major == 1 && g_dev.minor <= 3)
  {
  printf("A GPU with compute capability >= 2.x is required for CUDA Versions >=7.0\n\n");
    printf("See http://www.mersenne.ca/cudalucas.php for a list of cards\n\n");
  exit (2);
  }
#endif

#if CUDART_VERSION >= 9000  //CUDA >=9.0 Does not support compute_2.x
  if (g_dev.major == 2)
  {
  printf("A GPU with compute capability >= 3.x is required for CUDA Versions >=9.0\n\n");
    printf("See http://www.mersenne.ca/cudalucas.php for a list of cards\n\n");
  exit (2);
  }
#endif


  if (show_prop)
  {
  printf ("%s %d-bit build, compiled %s @ %s\n\n",    program,(int)(sizeof(void*)*8),__DATE__,__TIME__);
  
  printf("binary compiled for CUDA   %d.%d\n", CUDART_VERSION/1000, CUDART_VERSION%100); //|
  printf("CUDA runtime version       %d.%d\n", rt_ver/1000, rt_ver%100);          //>from mfaktc
  printf("CUDA driver version        %d.%d\n\n", drv_ver/1000, drv_ver%100);        //|
  printf ("---------------- DEVICE %d ----------------\n",    device_number);
  printf ("Device Name               %s\n",       g_dev.name);
  printf ("ECC Support?              %s\n",        g_dev.ECCEnabled ? "Enabled" : "Disabled");
  printf ("Compatibility             %d.%d\n",    g_dev.major, g_dev.minor);
  printf ("clockRate (MHz)           %d\n",       g_dev.clockRate/1000);
  printf ("memClockRate (MHz)        %d\n",       g_dev.memoryClockRate/1000);
  printf ("totalGlobalMem            %llu\n",     (unsigned long long) g_dev.totalGlobalMem);
  printf ("totalConstMem             %llu\n",     (unsigned long long) g_dev.totalConstMem);
  printf ("l2CacheSize               %d\n",       g_dev.l2CacheSize);
  printf ("sharedMemPerBlock         %llu\n",     (unsigned long long) g_dev.sharedMemPerBlock);
  printf ("regsPerBlock              %d\n",       g_dev.regsPerBlock);
  printf ("warpSize                  %d\n",       g_dev.warpSize);
  printf ("memPitch                  %llu\n",     (unsigned long long) g_dev.memPitch);
  printf ("maxThreadsPerBlock        %d\n",       g_dev.maxThreadsPerBlock);
  printf ("maxThreadsPerMP           %d\n",       g_dev.maxThreadsPerMultiProcessor);
  printf ("multiProcessorCount       %d\n",       g_dev.multiProcessorCount);
  printf ("maxThreadsDim[3]          %d,%d,%d\n", g_dev.maxThreadsDim[0], g_dev.maxThreadsDim[1], g_dev.maxThreadsDim[2]);
  printf ("maxGridSize[3]            %d,%d,%d\n", g_dev.maxGridSize[0], g_dev.maxGridSize[1], g_dev.maxGridSize[2]);
  printf ("textureAlignment          %llu\n",     (unsigned long long) g_dev.textureAlignment);
  printf ("deviceOverlap             %d\n",       g_dev.deviceOverlap);
  printf ("pciDeviceID               %d\n",       g_dev.pciDeviceID);
  printf ("pciBusID                  %d\n\n",     g_dev.pciBusID);
  }
}

int init_ffts(int new_len)
{
  #define COUNT 162
  FILE *fft;
  char buf[192];
  char fftfile[32];
  int next_fft, j = 0, i = 0, k = 1;
  int temp_fft[256] = {0};
  int default_mult[COUNT] = {  //this batch from GTX570 timings, at least up to 16384
                1, 2,  4,  8,  10,  14,    16,    18,    20,    32,    36,    42,
                 48,    50,    56,    60,    64,    70,    80,    84,    96,   112,
                120,   126,   128,   144,   160,   162,   168,   180,   192,   224,
                256,   288,   320,   324,   336,   360,   384,   392,   400,   448,
                512,   576,   640,   648,   672,   720,   768,   784,   800,   864,
                896,   900,  1024,  1152,  1176,  1280,  1296,  1344,  1440,  1568,
               1600,  1728,  1792,  2048,  2160,  2304,  2352,  2592,  2688,  2880,
               3024,  3136,  3200,  3584,  3600,  4096,  4320,  4608,  4704,  5120,
               5184,  5600,  5760,  6048,  6144,  6272,  6400,  6480,  7168,  7200,
               7776,  8064,  8192,  8640,  9216,  9408, 10240, 10368, 10584, 10800,
              11200, 11520, 12096, 12288, 12544, 12960, 13824, 14336, 14400, 16384,
              17496, 18144, 19208, 19600, 20000, 20250, 21952, 23328, 23814, 24300,
              24500, 25088, 25600, 26244, 27000, 27216, 28000, 28672, 31104, 31250,
              32000, 32400, 32768, 33614, 34992, 36000, 36288, 38416, 39200, 39366,
              40500, 41472, 42336, 43200, 43904, 47628, 49000, 50000, 50176, 51200,
              52488, 54432, 55296, 56000, 57344, 60750, 62500, 64000, 64800, 65536};


  //read fft.txt into temp array, inserting new_len appropriately
  sprintf (fftfile, "%s fft.txt", g_dev.name);
  fft = fopen_and_lock(fftfile, "r");
  if(fft)
  {
  while(fgets(buf, 192, fft) != NULL && i < 254)
  {
    if(next_fft = atoi(buf))
    {
    next_fft <<= 10;
    if((new_len > temp_fft[i]) && (new_len < next_fft))
    {
      i++;
      temp_fft[i] = new_len;
    }
    i++;
    temp_fft[i] = next_fft;
    }
  }
  unlock_and_fclose(fft);
  }
  else
  {
  i = 1;
  temp_fft[1] = new_len;
  }
  //put default values less than smallest entry in fft.txt into fft array
  while((j < COUNT) && (1024 * default_mult[j] < temp_fft[1]))
  {
  g_ffts[j] = default_mult[j] << 10;
  j++;
  }
  //put saved fft.txt values into fft array
  while(k <= i && j < 256)
  {
  g_ffts[j] = temp_fft[k];
  k++;
  j++;
  }
  k = 0;
  if (j) while(k < COUNT && default_mult[k] * 1024 <= g_ffts[j - 1]) k++;
  //put default values bigger than largest entry in fft.txt into fft array
  while(k < COUNT && j < 256)
  {
  g_ffts[j] = default_mult[k] << 10;
  j++;
  k++;
  }

  return j;
}

int choose_fft_length (int q, int *index)
{
/* In order to increase length if an exponent has a round off issue, we use an
extra paramter that we can adjust on the fly. In check(), index starts as -1,
the default. In that case, choose from the table. If index >= 0, we must assume
it's an override index and return the corresponding length. If index > table-count,
then we assume it's a manual fftlen and return the proper index. */
  double ff1 = 0.999 * 1024.0 * 0.0000357822505975293;
  double ff2 = 0.99 * 1024.0 * 0.0002670641830112380;
  double e1 = 1.022179977969700;
  double e2 = 0.929905288591965;
  int lb;
  int ub;
  int i = 0;

  if(q > 0)
  {
    lb = (int) ceil (ff1 * exp (e1 * log ((double) q)));
    ub = (int) floor(ff2 * exp (e2 * log ((double) q)));
    g_lbi = 0;
    while( g_lbi < g_fft_count && g_ffts[g_lbi] < lb) g_lbi++;
    g_ubi = g_lbi;
    while( g_ubi < g_fft_count && g_ffts[g_ubi] <= ub) g_ubi++;
    g_ubi--;
  }
  //printf("Index: %d, Lower bound at %d: %dK, Upper bound at %d: %dK\n", *index, g_lbi, g_ffts[g_lbi]/1024, g_ubi, g_ffts[g_ubi]/1024);

  if(*index >= g_fft_count) while(i < g_fft_count && g_ffts[i] < *index) i++;
  else i = *index;

  if(i < g_lbi)
  {
    if(*index) printf("The fft length %dK is too small for exponent %d, increasing to %dK\n", g_ffts[i] / 1024, q, g_ffts[g_lbi] / 1024);
    i = g_lbi;
  }
  if(i > g_ubi)
  {
    printf("The fft length %dK is too large for exponent %d, decreasing to %dK\n", g_ffts[i] / 1024, q, g_ffts[g_ubi] / 1024);
    i = g_ubi;
  }
  *index = i;

  return g_ffts[i];
}

int fft_from_str(const char* str)
/* This is really just strtoul with some extra magic to deal with K or M */
{
  char* endptr;
  const char* ptr = str;
  int len, mult = 1;
  while( *ptr ) {
  if( *ptr == 'k' || *ptr == 'K' ) {
    mult = 1024;
    break;
  }
  if( *ptr == 'm' || *ptr == 'M' ) {
    mult = 1024*1024;
    break;
  }
  ptr++;
  }

  len = (int) strtoul(str, &endptr, 10)*mult;
  if( endptr != ptr ) { // The K or M must directly follow the num (or the num must extend to the end of the str)
  fprintf (stderr, "can't parse fft length \"%s\"\n\n", str);
  exit (2);
  }
  return len;
}

void alloc_gpu_mem(int n)
{
  cutilSafeCall (cudaMalloc ((void **) &d_x, sizeof (cufftDoubleComplex) * n/2));
  h_x_debug = (cufftDoubleComplex *) malloc(sizeof (cufftDoubleComplex) * n/2);
  cutilSafeCall (cudaMalloc ((void **) &d_negacyclic_weight, sizeof (cufftDoubleComplex) * n/2));
  cutilSafeCall (cudaMalloc ((void **) &d_irrational_weight, sizeof (double2) * n/2));
  cutilSafeCall (cudaMalloc ((void **) &d_irrational_weight_inverse, sizeof (double2) * n/2));

  cutilSafeCall (cudaMalloc ((void **) &d_xint, sizeof (int2) * n/2));
  h_xint_debug = (int2 *) malloc(sizeof(int2) * n/2);
  //h_xint_simulate = (int2 *) malloc(sizeof(int2) * n/2);
  cutilSafeCall (cudaMalloc ((void **) &d_base_bit, sizeof (char2) * n/2));

  cutilSafeCall (cudaMalloc ((void **) &d_block_carry, sizeof (int2) * n/2/128));
  h_bc_debug = (int2 *) malloc(sizeof(int2) * n/2/128);
  //h_bc_simulate = (longlong2 *) malloc(sizeof(longlong2) * n/2/128);
}

void write_gpu_data(int q, int n)
{
  int j;

  float zero = 0.0;
  cudaMemcpyToSymbol (g_err, &zero, sizeof (float));

  double inv_n = 1.0 / (double) n;

  // Negacyclic weight
  cufftDoubleComplex *h_negacyclic_weight = (cufftDoubleComplex *) malloc(sizeof(cufftDoubleComplex) * n/2);
  h_negacyclic_weight[0].x = 1.0;
  h_negacyclic_weight[0].y = 0.0;
  h_negacyclic_weight[n/4].x = h_negacyclic_weight[n/4].y = M_SQRT1_2;
  for(j = 1; j < n/4; j++)
  {
    double c = cospi(j * inv_n);
    double s = sinpi(j * inv_n);

    h_negacyclic_weight[j].x = c;
    h_negacyclic_weight[j].y = s;
    h_negacyclic_weight[n/2-j].x = s;
    h_negacyclic_weight[n/2-j].y = c;
  }
  cutilSafeCall (cudaMemcpy (d_negacyclic_weight, h_negacyclic_weight, sizeof (cufftDoubleComplex) * n/2, cudaMemcpyHostToDevice));

  
  // irrational base weight -- forward & inverse
  double2 *h_irrational_weight = (double2 *) malloc(sizeof(double2) * n/2);
  double2 *h_irrational_weight_inverse = (double2 *) malloc(sizeof(double2) * n/2);
  int r = 0;
  const int qmod = q % n;
  double w;

  for(j = 0; j < n; j++)
  {
    // jth weight = ceil(q*j/N)- (q*j/N) = 1 - (q*j % N) / N
  
    w = r ? exp2((n-r) * inv_n) : 1.0;  // 1.0 - r*inv_n = (n-r) * inv_n
   
    if(j < n/2) // stuff lower half into real component
    {
      h_irrational_weight[j].x = w;
      h_irrational_weight_inverse[j].x = inv_n*2.0/w;
    }
    else // upper half into imaginary component
    {
      h_irrational_weight[j-n/2].y = w;
      h_irrational_weight_inverse[j-n/2].y = inv_n*2.0/w;
    }
    r += qmod;
    if(r >= n) r -= n;
  }
  cutilSafeCall (cudaMemcpy (d_irrational_weight, h_irrational_weight, sizeof (double2) * n/2, cudaMemcpyHostToDevice));
  cutilSafeCall (cudaMemcpy (d_irrational_weight_inverse, h_irrational_weight_inverse, sizeof (double2) * n/2, cudaMemcpyHostToDevice));

  // # of data bits per fft limb
  h_base_bit = (char2 *) malloc (sizeof (char2) * n/2);

  // jth limb size = ceil(q*(j+1)/n) - ceil(q*j/n) bits

  const int qby = q/n;
  int quot0 = 0; // (q*0) / n
  int rem0 = 0; // (q*0) % n
  int quot1, rem1;

  for(j = 0; j < n; j++)
  {
    quot1 = quot0 + qby; rem1 = rem0 + qmod;
    if(rem1 >= n)
    {
      rem1 -= n;
      quot1++;
    }
    int q1_ceil = quot1 + (rem1 ? 1 : 0);
    int q0_ceil = quot0 + (rem0 ? 1 : 0);
    if(j < n/2)
    {
      h_base_bit[j].x = (char) (q1_ceil - q0_ceil);
    }
    else
    {
      h_base_bit[j-n/2].y = (char) (q1_ceil - q0_ceil);
    }
    quot0 = quot1;
    rem0 = rem1;
  }
  cudaMemcpy (d_base_bit, h_base_bit, sizeof (char2) * n/2, cudaMemcpyHostToDevice);

  cufftSafeCall (cufftPlan1d (&g_plan, n/2, CUFFT_Z2Z, 1));


  free ((char *) h_irrational_weight_inverse);
  free ((char *) h_irrational_weight);
  free ((char *) h_negacyclic_weight);
}

void init_x(int2 *x_int, unsigned *x_packed, int q, int n, int new_test)
{
  // xpacked == NULL means this is being called from cufftbench
  int j;
  const int end = (q + 31) / 32;
  const int width = end + 10;

  if(new_test)
  {
    for(j = 0; j <  n/2; j++)
    {
      x_int[j].x = 0; x_int[j].y = 0; 
      if(x_packed)
      {
        gec.h_cumulative_int[j].x = 0; gec.h_cumulative_int[j].y = 0;
        gec.h_base_int[j].x = 0; gec.h_base_int[j].y = 0; 
      }
    }

    x_int[0].x = 3; // PRP testing base. should be parameterized. for now harcoded to 3
    if(x_packed) gec.h_cumulative_int[0].x = 3;
    if(x_packed) gec.h_base_int[0].x = 3;

    if(x_packed)
    {
      for(j = 0; j < end; j++)
      {
        x_packed[j] = 0;
        x_packed[j+width] = 0;
        x_packed[j+width*2] = 0;
      }
      x_packed[0] = 3; // same here. hardcoded prp base
      x_packed[end] = q;
      x_packed[end + 1] = n;
      x_packed[0+width] = 3; 
      x_packed[end+width] = q;
      x_packed[end + 1+width] = n;
      x_packed[0+width*2] = 3;
      x_packed[end+width*2] = q;
      x_packed[end + 1+width*2] = n;
    }
  }

  cudaMemcpy (d_xint, x_int, sizeof (int2) * n/2 , cudaMemcpyHostToDevice);
  weight <<< n/2/128, 128 >>> (n, d_xint, d_x, d_negacyclic_weight, d_irrational_weight);

  if(x_packed)
  {
    cudaMemcpy (gec.d_cumulative_int, gec.h_base_int, sizeof (int2) * n/2 , cudaMemcpyHostToDevice);
    weight <<< n/2/128, 128 >>> (n, gec.d_cumulative_int, gec.d_base, d_negacyclic_weight, d_irrational_weight);

    cudaMemcpy (gec.d_cumulative_int, gec.h_cumulative_int, sizeof (int2) * n/2 , cudaMemcpyHostToDevice);
    weight <<< n/2/128, 128 >>> (n, gec.d_cumulative_int, gec.d_cumulative, d_negacyclic_weight, d_irrational_weight);
  }
}

void unpack_bits_and_balance_digits(int2 *x_int, unsigned *packed_x, int q , int n)
{
  int bits_in_accumulator = 0;
  unsigned long long accumulator = 0;
  int curr = 0;
  int carry = 0;

  for(int i = 0; i < n; i++)
  {
    int bits_to_grab = (i < n/2) ? h_base_bit[i].x : h_base_bit[i-n/2].y;

    if (bits_in_accumulator < bits_to_grab) { // not enough bits. get more.
      accumulator |= ((unsigned long long) packed_x[curr]) << bits_in_accumulator;
      bits_in_accumulator += 32;
      curr++;
    }

    unsigned base = 1 << bits_to_grab;
    unsigned mask = base-1;
    unsigned cmp  = base >> 1;
 
    int the_bits = (unsigned) accumulator & mask;
    accumulator >>= bits_to_grab;
    bits_in_accumulator -= bits_to_grab;

    the_bits += carry; // carry in from the previous digit's balancing
    carry = (the_bits >= (int) cmp) ? 1 : 0; // new carry out
    if( carry )
      the_bits -= base;

    if(i < n/2)
    {
      x_int[i].x = the_bits;
    }
    else
    {
      x_int[i-n/2].y = the_bits;
    }
  }

  x_int[0].x -= carry; // carry out from highest limb causes the lowest one to be subtracted instead of added
}

void init_gec(int q, int n); // forward declaration
void free_gec(); // forward declaration

int2 *init_lucas(unsigned *x_packed,
               int       q,
               int       &n,
               int       &j,
               unsigned long long &total_time,
               unsigned long long &time_adj,
               unsigned  &iter_adj)
{
  int2 *x_int;
  const int end = (q + 31) / 32;
  const int width = end + 10;

  int new_test = 0;

  n = x_packed[end + 1];
  if(n == 0) new_test = 1;

  j = x_packed[end + 2];
  total_time = (unsigned long long) x_packed[end + 4] << 15;
  time_adj = (unsigned long long) x_packed[end + 5] << 15;
  iter_adj = x_packed[end + 6];

  if(g_fftlen == 0) g_fftlen = n;
  if(g_fft_count == 0) g_fft_count = init_ffts(g_fftlen);
  n = choose_fft_length(q, &g_fftlen);

  if(n != (int) x_packed[end + 1])
  {
    x_packed[end + 1] = n;
    time_adj = total_time;
    if(j > 1) iter_adj = j;
  }

  x_int = (int2 *) malloc (sizeof (int2) * n/2);
  gec.rollback_count_total = (int) x_packed[end + 3];
  gec.last_cumulative_iteration = (x_packed+width)[end + 2];
  gec.accumulate_count = (x_packed+width)[end + 3];
  gec.last_check_iteration = (x_packed+width)[end + 4];
  init_gec(q, n);

  alloc_gpu_mem(n);
  write_gpu_data(q, n);
  if(!new_test)
  {
    unpack_bits_and_balance_digits(x_int, x_packed, q, n);
    unpack_bits_and_balance_digits(gec.h_cumulative_int, x_packed+width, q, n);
    unpack_bits_and_balance_digits(gec.h_base_int, x_packed+width*2, q, n);
  }
  init_x(x_int, x_packed, q, n, new_test);
  return x_int;
}

/**************************************************************************
 *                                                                        *
 *                               Cleanup                                  *
 *                                                                        *
 **************************************************************************/

void free_host (int2 *x)
{
  free ((char *) h_base_bit);
  h_base_bit = NULL;

  if(x) free ((char *) x);
}

void free_gpu(int dp)
{
  if(dp) cufftSafeCall (cufftDestroy (g_plan));
  cutilSafeCall (cudaFree ((char *) d_x));
  cutilSafeCall (cudaFree ((char *) d_negacyclic_weight));
  cutilSafeCall (cudaFree ((char *) d_xint));
  cutilSafeCall (cudaFree ((char *) d_block_carry));
  free((char *) h_x_debug);
  free((char *) h_xint_debug);
  //free((char *) h_xint_simulate);
  free((char *) h_bc_debug);
  //free((char *) h_bc_simulate);
}

void close_test (int2 *x)
{
  free_host(x);
  free_gpu(1);
  free_gec();
}

/**************************************************************************
 *                                                                        *
 *                        Checkpoint Processing                           *
 *                                                                        *
 **************************************************************************/


//From apsen
void print_time_from_seconds (unsigned in_sec, char *res, int mode)
{
  char s_time[32] = {0};
  unsigned day = 0;
  unsigned hour;
  unsigned min;
  unsigned sec;

  if(mode >= 2)
  {
    day = in_sec / 86400;
    in_sec %= 86400;
  }
  hour = in_sec / 3600;
  in_sec %= 3600;
  min = in_sec / 60;
  sec = in_sec % 60;
  if (day)
  {
    if(mode & 1) sprintf (s_time, "%ud%02uh%02um%02us", day, hour, min, sec);
    else sprintf (s_time, "%u:%02u:%02u:%02u", day, hour, min, sec);
  }
  else if (hour)
  {
    if(mode & 1) sprintf (s_time, "%uh%02um%02us", hour, min, sec);
    else sprintf (s_time, "%u:%02u:%02u", hour, min, sec);
  }
  else
  {
    if(mode & 1) sprintf (s_time, "%um%02us", min, sec);
    else sprintf (s_time, "%u:%02u", min, sec);
  }
  if(res) sprintf(res, "%12s", s_time);
  else printf ("%s", s_time);
}

int standardize_digits(int2 *x_int, int q, int n)
{
  int i, carry = 0;

  for(i = 0; i < n; i++)
  {
    int val = (i < n/2) ? x_int[i].x : x_int[i-n/2].y;
    val += carry;

    int bits = (i < n/2) ? h_base_bit[i].x : h_base_bit[i-n/2].y;
    int base = 1 << bits;

    carry = 0;
    while (val >=base) { val -= base; carry++; }
    while (val < 0) { val += base; carry--; }

    if(i < n/2)
    {
      x_int[i].x = val;
    }
    else
    {
      x_int[i-n/2].y = val;
    }
  }

  if( carry )
  {
    x_int[0].x -= carry;
    return standardize_digits(x_int, q, n); // too lazy.
  }
  else
    return(0);
}

unsigned long long find_residue(unsigned *packed_x, int q, int n)
{
  unsigned long long residue = packed_x[1];

  residue = (residue << 32) | packed_x[0];

  return residue;
}

void printbits(unsigned long long residue, int q, int n, FILE* fp, unsigned * packed_bits = NULL)
{
  if (fp)
  {
    printf ("W( %d )C, 0x%016llx,", q, residue);
    fprintf (fp, "W( %d )C, 0x%016llx,", q, residue);
    fprintf (fp, " n = %dK, %s", n/1024, program);
    if(packed_bits) // RES2048
    {
      fprintf(fp, ", RES2048: 0x");
      for(int i=2048/32-1; i>=0; i--)
      {
        fprintf(fp, "%08X", packed_bits[i]);
      }
    }

    fprintf(fp, ", gec_errors: %d", gec.rollback_count_total);
  }
  else printf ("/ %d, 0x%016llx,", q, residue);
  printf (" %dK, %s", n/1024, program);
}

void pack_bits(int2 *x_int, unsigned *packed_x, int q , int n)
{
  unsigned long long val, accumulator = 0;
  int i, j = 0, bits_in_accumulator = 0;

  for(i = 0; i < n; i++)
  {
    val = (unsigned) ((i < n/2) ? x_int[i].x : x_int[i-n/2].y);
    accumulator |= (val << bits_in_accumulator);
    bits_in_accumulator += ((i < n/2) ? h_base_bit[i].x : h_base_bit[i-n/2].y);
    if(bits_in_accumulator >= 32)
    {
      packed_x[j] = (unsigned) accumulator;
      accumulator >>= 32;
      bits_in_accumulator -= 32;
      j++;
    }
  }
  packed_x[j] = (unsigned) accumulator;
}

void set_checkpoint_data(unsigned *x_packed, int q, int j, unsigned long long tt, unsigned long long ta, unsigned ia)
{
  int end = (q + 31) / 32;

  x_packed[end + 2] = j;
  x_packed[end + 3] = gec.rollback_count_total;
  x_packed[end + 4] = tt >> 15;
  x_packed[end + 5] = ta >> 15;
  x_packed[end + 6] = ia;
}

void reset_err()
{
  float zero = 0.0;

  cudaMemcpyToSymbol (g_err, &zero, sizeof (float));
}

void process_output(int q,
          int n,
          int j,
          int last_chk,
          float maxerr,
          unsigned long long residue,
          unsigned long long diff,
          unsigned long long diff1,
          unsigned long long diff2,
          unsigned long long total_time)
{
  time_t now;
  struct tm *tm_now = NULL;
  int index = 0;
  char buffer[192];
  char temp [32];
  int i;
  static int header_iter = 0;

  now = time(NULL);
  tm_now = localtime(&now);
  strftime(g_output[0], 32, "%a", tm_now);                             //day of the week Sat
  strftime(g_output[1], 32, "%b", tm_now);                             //month
  strftime(g_output[2], 32, "%d", tm_now);                             //date
  strftime(g_output[3], 32, "%H", tm_now);                             //hour
  strftime(g_output[4], 32, ":%M", tm_now);                            //min
  strftime(g_output[5], 32, ":%S", tm_now);                            //sec
  strftime(g_output[6], 32, "%y", tm_now);                             //year
  strftime(g_output[7], 32, "%Z", tm_now);                             //time zone
  sprintf(g_output[8],  "%9d", q);                                      //exponent
  sprintf(temp,  "W%d", q);
  sprintf(g_output[9],  "%10s", temp);                                 //Mersenne number
  sprintf(g_output[10], "%5dK", n / 1024);                             //fft, multiple of 1k
  sprintf(g_output[11], "%8d", n);                                     //fft
  sprintf(g_output[12], "%9d", j);                                     //iteration
  sprintf(g_output[13], "%9d", 0);                                     //offset
  sprintf(g_output[14], "0x%016llx", residue);                         //residue, lower case hex
  sprintf(g_output[15], "0x%016llX", residue);                         //residue, upper case hex
  sprintf(g_output[18], "%s", program);                                //program name
  sprintf(g_output[19], "%%");                                         //percent done
  sprintf(g_output[25], "%7.5f", maxerr);                              //round off error
  sprintf(g_output[26], "%9.5f", diff1 / 1000.0 / (j - last_chk));     //ms/iter
  sprintf(g_output[27], "%10.5f", diff1 / 1000000.0 );                 //time since last checkpoint xxx.xxxxx seconds format
  sprintf(g_output[28], "%10.5f", (j - last_chk) * 1000000.0 / diff1); //iter/sec
  sprintf(g_output[29], "%10.5f", j / (float) (q - 2) * 100 );         //percent done
  print_time_from_seconds((unsigned) diff, g_output[16], 0);                      //time since last checkpoint hh:mm:ss format
  print_time_from_seconds((unsigned) diff2, g_output[17], 2);                     //ETA in d:hh:mm:ss format

  i = 0;
  while( i < 50 && g_output_code[i] >= 0)
  {
    index += sprintf(buffer + index,"%s",g_output[g_output_code[i]]);
    if(g_output_code[i] >= 25 && g_output_code[i] < 30)
    {
      i++;
      index -= g_output_code[i];
    }
    i++;
  }

  if(header_iter == 0 && g_output_interval > 0)
  {
    header_iter = g_output_interval - 1;
    printf("%s\n", g_output_header);
  }
  header_iter--;
  printf("%s\n", buffer);
}

/**************************************************************************
 *                                                                        *
 *                        File Related Functions                          *
 *                                                                        *
 **************************************************************************/
unsigned magic_number(unsigned *x_packed, int q)
{
  return 0xEa657aff; // looks like WAGSTAFF
}

unsigned int checkpoint_checksum(char *string, int chars)
/* generates a CRC-32 like checksum of the string */
{
  unsigned int chksum=0;
  int i,j;

  for(i=0;i<chars;i++)
  {
  for(j=7;j>=0;j--)
  {
    if((chksum>>31) == (((unsigned int)(string[i]>>j))&1))
    {
    chksum<<=1;
    }
    else
    {
    chksum = (chksum<<1)^0x04C11DB7;
    }
  }
  }
  return chksum;
}

unsigned ch_sum(unsigned *x_packed, int q)
{
  int end = (q + 31) / 32;
  int j;
  unsigned sum = 0;

  for(j = 0; j < end + 9; j++) sum += x_packed[j];
  return sum;
}


unsigned *read_checkpoint(int q)
{
  FILE *fPtr;
  unsigned *x_packed;
  char chkpnt_cfn[32];
  char chkpnt_tfn[32];
  const int end = (q + 31) / 32;
  const int width = end + 10;
  const int check_point_size = sizeof (unsigned) * width * 3; 
  // 3 values in 1
  x_packed = (unsigned *) malloc (check_point_size); // 1. the actual last iteration
  unsigned * x_packed2 = x_packed + width; // 2. the running product of GEC
  unsigned * x_packed3 = x_packed2 + width; // 3. the base value of GEC

  sprintf (chkpnt_cfn, "cW%d", q);
  sprintf (chkpnt_tfn, "tW%d", q);

  size_t bytes_read;

/* Attempt reading from main checkpoint cW */
  fPtr = fopen (chkpnt_cfn, "rb");
  if (!fPtr) goto process_backup;

  bytes_read = fread (x_packed, 1,  check_point_size, fPtr);
  fclose (fPtr);

  if(bytes_read != check_point_size)
  {
    fprintf (stderr, "\nThe checkpoint appears to be corrupt. Trying the backup file.\n");
    goto process_backup;
  }

  if(x_packed [width-1] != checkpoint_checksum((char*) x_packed , sizeof (unsigned) * (width-1))
  || x_packed2[width-1] != checkpoint_checksum((char*) x_packed2, sizeof (unsigned) * (width-1))
  || x_packed3[width-1] != checkpoint_checksum((char*) x_packed3, sizeof (unsigned) * (width-1))
    )
  {
    fprintf (stderr, "\nThe checkpoint appears to be corrupt. Trying the backup file.\n");
    goto process_backup;
  }

  return x_packed;


process_backup:
/* Attempt reading from backup checkpoint tW */
  fPtr = fopen(chkpnt_tfn, "rb");
  if (!fPtr) goto return_empty;

  bytes_read = fread (x_packed, 1, check_point_size, fPtr);
  fclose (fPtr);

  if(bytes_read != check_point_size)
  {
    fprintf (stderr, "\nThe backup appears to be corrupt. Restarting test.\n");
    goto return_empty;
  }

  if(x_packed [width-1] != checkpoint_checksum((char*) x_packed , sizeof (unsigned) * (width-1))
  || x_packed2[width-1] != checkpoint_checksum((char*) x_packed2, sizeof (unsigned) * (width-1))
  || x_packed3[width-1] != checkpoint_checksum((char*) x_packed3, sizeof (unsigned) * (width-1))
    )
  {
    fprintf (stderr, "\nThe backup appears to be corrupt. Restarting test.\n");
    goto return_empty;
  }

  return x_packed;

return_empty:
  memset ( (void *) x_packed, 0, check_point_size );
  x_packed[end + 2] = 1;
  x_packed[end + 3] = (unsigned) -1;
  return x_packed;
}

void write_checkpoint(unsigned *x_packed, int q, unsigned long long residue)
{
  FILE *fPtr;
  char chkpnt_cfn[32];
  char chkpnt_tfn[32];

  const int end = (q + 31) / 32;
  const int width = end + 10;
  const int check_point_size = sizeof (unsigned) * width * 3; 

  sprintf (chkpnt_cfn, "cW%d", q);
  sprintf (chkpnt_tfn, "tW%d", q);
  (void) unlink (chkpnt_tfn);
  (void) rename (chkpnt_cfn, chkpnt_tfn);
  fPtr = fopen (chkpnt_cfn, "wb");
  if (!fPtr)
  {
    fprintf(stderr, "Couldn't write checkpoint.\n");
    return;
  }
  x_packed[end + 8] = magic_number(x_packed, q);
  x_packed[end + 9] = checkpoint_checksum((char*) x_packed, sizeof (unsigned) * (width-1));
  (x_packed+width)[end + 8] = magic_number(x_packed+width, q);
  (x_packed+width)[end + 9] = checkpoint_checksum((char*) (x_packed+width), sizeof (unsigned) * (width-1));
  (x_packed+width*2)[end + 8] = magic_number(x_packed+width*2, q);
  (x_packed+width*2)[end + 9] = checkpoint_checksum((char*) (x_packed+width*2), sizeof (unsigned) * (width-1));

  fwrite (x_packed, 1, check_point_size, fPtr);
  fclose (fPtr);
}

void rm_checkpoint (int q)
{
  char chkpnt_cfn[32];
  char chkpnt_tfn[32];
  sprintf (chkpnt_cfn, "cW%d", q);
  sprintf (chkpnt_tfn, "tW%d", q);
  (void) unlink (chkpnt_cfn);
  (void) unlink (chkpnt_tfn);
}

// GEC check
/* The math. L1 = stride. L2=howmany L1 strides before we do full check (i.e L1*L2 iterations).
ideally, L1=L2, but it should be fine if L1 ~= L2. This allows more flexibility towards the end.
Default L1=L2=1000

u(0)=3
u(t)=3^(2^t) mod N  [0]
d(t)=u(0)*u(L1)*u(2*L1)*...*u(t*L1) mod N  [1]

d(L2)=d(L2-1)*u(L2*L1) mod N  [2]
d(L2)=u(0)*d(L2-1)^(2^L1) mod N  [3]

gec.d_base = u(0)
gec.d_cumulative = d(t)
gec.d_scratch_cumulative = d(L2-1)
*/

int gec_accumulate_and_check (int q, int n, int iters)
{
  gec.accumulate_count++; // increase the count

// cost of accumulate step = 2 FFT. the d_x one doesn't count because we will reuse that in the next prp_square iteration

  cufftExecZ2Z (g_plan, gec.d_cumulative, gec.d_cumulative, CUFFT_FORWARD);
  cufftExecZ2Z (g_plan, d_x, gec.d_x_scratch, CUFFT_FORWARD);
  // before doing the full check, take a backup of the previous value from gec.d_cumulative into gec.d_scratch_cumulative
  // controlled by gec.accumulate_count == L2 condition
  multiply <<< n / 2 / 128, 128 >>> (n, gec.d_cumulative, gec.d_x_scratch, gec.accumulate_count == gec.L2, gec.d_scratch_cumulative); 
  cufftExecZ2Z (g_plan, gec.d_cumulative, gec.d_cumulative, CUFFT_INVERSE);
  unweight_and_partial_carry_unsafe_opt <<< n / 2 / 128, 128 >>> (n, gec.d_cumulative, d_negacyclic_weight, d_irrational_weight_inverse, gec.d_cumulative_int, d_base_bit, d_block_carry);
  finish_carry_and_reweight <<< n / 2 / 128, 128 >>> (n, gec.d_cumulative, d_negacyclic_weight, d_irrational_weight, gec.d_cumulative_int, d_base_bit, d_block_carry);

  gec.last_cumulative_iteration = iters;

  if(gec.accumulate_count < gec.L2) return -1; // that's it for now. nothing further to do.

  // copy for comparison
  cutilSafeCall (cudaMemcpy (gec.h_cumulative_int, gec.d_cumulative_int, sizeof (int2) * n/2, cudaMemcpyDeviceToHost));
  standardize_digits(gec.h_cumulative_int, q, n);

// cost of square step = 2*L1-1 FFT. first one is free.
  for(int i=0; i<gec.L1; i++)
  {
    if(i>0) cufftExecZ2Z (g_plan, gec.d_scratch_cumulative, gec.d_scratch_cumulative, CUFFT_FORWARD); // when entering the loop gec.d_scratch_cumulative is already forward transformed.
    square <<< n / 2 / 128, 128 >>> (n, gec.d_scratch_cumulative);
    cufftExecZ2Z (g_plan, gec.d_scratch_cumulative, gec.d_scratch_cumulative, CUFFT_INVERSE);
    unweight_and_partial_carry_unsafe_opt <<< n / 2 / 128, 128 >>> (n, gec.d_scratch_cumulative, d_negacyclic_weight, d_irrational_weight_inverse, gec.d_cumulative_int, d_base_bit, d_block_carry);
    finish_carry_and_reweight <<< n / 2 / 128, 128 >>> (n, gec.d_scratch_cumulative, d_negacyclic_weight, d_irrational_weight, gec.d_cumulative_int, d_base_bit, d_block_carry);
  }

// plus 3
  cufftExecZ2Z (g_plan, gec.d_scratch_cumulative, gec.d_scratch_cumulative, CUFFT_FORWARD);
  cufftExecZ2Z (g_plan, gec.d_base, gec.d_base_scratch, CUFFT_FORWARD);
  multiply <<< n / 2 / 128, 128 >>> (n, gec.d_scratch_cumulative, gec.d_base_scratch, 0, NULL); // d_scratch_cumulative *= d_base_scratch
  cufftExecZ2Z (g_plan, gec.d_scratch_cumulative, gec.d_scratch_cumulative, CUFFT_INVERSE);
  unweight_and_partial_carry_unsafe_opt <<< n / 2 / 128, 128 >>> (n, gec.d_scratch_cumulative, d_negacyclic_weight, d_irrational_weight_inverse, gec.d_cumulative_int, d_base_bit, d_block_carry);
  finish_carry_and_reweight <<< n / 2 / 128, 128 >>> (n, gec.d_scratch_cumulative, d_negacyclic_weight, d_irrational_weight, gec.d_cumulative_int, d_base_bit, d_block_carry);

  cutilSafeCall (cudaMemcpy (gec.h_cumulative_int_scratch, gec.d_cumulative_int, sizeof (int2) * n/2, cudaMemcpyDeviceToHost));
  standardize_digits(gec.h_cumulative_int_scratch, q, n);

  int flag = 1;
  for(int i=0; i<n/2; i++)
  {
    if(gec.h_cumulative_int[i].x != gec.h_cumulative_int_scratch[i].x || gec.h_cumulative_int[i].y != gec.h_cumulative_int_scratch[i].y)
    {
      flag = 0;
      break;
    }
  }

  if(flag) gec.last_check_iteration = iters; // update only if success

  return flag;
}

void pick_L1_L2(int iterations_left, int &L1, int &L2)
{
  // iterations_left must be >= 100

  L1 = 1000; L2 = 1000;

  if(L1 * L2 <= iterations_left) return;

  // pick a L1/L2 combination near sqrt(iterations_left) that minimizes (iterations_left - L1*L2)
  int sq = (int) sqrtf( (float) iterations_left ); // sq >= 10

  int min_val = 2000;
  int delta = sq / 8;
  for(int i=sq-delta; i<=sq+delta; i++) // don't go too far off the center. the further away from sq, the less efficient GEC becomes
  {
    int l1 = i, l2 = iterations_left / l1;
    int iters_not_covered = iterations_left - l1*l2;
    if(iters_not_covered < min_val)
    {
      L1 = l1; L2 = l2; min_val = iters_not_covered;
    }
  }
}

void init_gec(int q, int n)
{
  int L1, L2;
  pick_L1_L2( q-gec.last_check_iteration, L1, L2 );
  
  gec.L1 = L1; gec.L2 = L2;
  printf("GEC L1 = %d L2 = %d\n", gec.L1, gec.L2);

  gec.rollback_count_this_pass = 0;

  cutilSafeCall (cudaMalloc ((void **) &gec.d_base, sizeof (cufftDoubleComplex) * n/2));
  cutilSafeCall (cudaMalloc ((void **) &gec.d_base_scratch, sizeof (cufftDoubleComplex) * n/2));
  cutilSafeCall (cudaMalloc ((void **) &gec.d_cumulative, sizeof (cufftDoubleComplex) * n/2));
  cutilSafeCall (cudaMalloc ((void **) &gec.d_scratch_cumulative, sizeof (cufftDoubleComplex) * n/2));
  cutilSafeCall (cudaMalloc ((void **) &gec.d_cumulative_int, sizeof (int2) * n/2));
  cutilSafeCall (cudaMalloc ((void **) &gec.d_x_scratch, sizeof (cufftDoubleComplex) * n/2));
  gec.h_cumulative_int = (int2 *) malloc(sizeof (int2) * n/2);
  gec.h_cumulative_int_scratch = (int2 *) malloc(sizeof (int2) * n/2);
  gec.h_base_int = (int2 *) malloc(sizeof (int2) * n/2);
}

void next_gec(int q, int n)
{
  int L1, L2;
  pick_L1_L2( q-gec.last_check_iteration, L1, L2 );
  
  gec.L1 = L1; gec.L2 = L2;
  printf("Next GEC L1 = %d L2 = %d\n", gec.L1, gec.L2);
  gec.accumulate_count = 0;
  gec.rollback_count_this_pass = 0; // made progress. so this can be reset.

  // copy the last iteration value as the new base and cumulative
  cutilSafeCall (cudaMemcpy (gec.d_base, d_x, sizeof (cufftDoubleComplex) * n/2, cudaMemcpyDeviceToDevice));
  cutilSafeCall (cudaMemcpy (gec.d_cumulative, d_x, sizeof (cufftDoubleComplex) * n/2, cudaMemcpyDeviceToDevice));
  cutilSafeCall (cudaMemcpy (gec.h_cumulative_int, d_xint, sizeof (int2) * n/2, cudaMemcpyDeviceToHost));
  memcpy( (void *) gec.h_base_int, (void *) gec.h_cumulative_int, sizeof(int2) * n/2 );
}

void free_gec()
{
  cutilSafeCall (cudaFree ((char *) gec.d_base));
  cutilSafeCall (cudaFree ((char *) gec.d_base_scratch));
  cutilSafeCall (cudaFree ((char *) gec.d_cumulative));
  cutilSafeCall (cudaFree ((char *) gec.d_scratch_cumulative));
  cutilSafeCall (cudaFree ((char *) gec.d_cumulative_int));
  cutilSafeCall (cudaFree ((char *) gec.d_x_scratch));
  free((char *) gec.h_cumulative_int);
  gec.h_cumulative_int = NULL;
  free((char *) gec.h_cumulative_int_scratch);
  gec.h_cumulative_int_scratch = NULL;
  free((char *) gec.h_base_int);
  gec.h_base_int = NULL;
}

void gec_rollback(int q, int n)
{
  gec.rollback_count_total++;
  gec.rollback_count_this_pass++;
  
  if( gec.rollback_count_this_pass > 3 )
  {
    printf("GEC Unable to make progress for 3 tries. Aborting...\n");
    exit(2);
  }

  gec.accumulate_count = 0;
  gec.last_cumulative_iteration = gec.last_check_iteration;


  // copy the last checked value as the new base. 
  cutilSafeCall (cudaMemcpy (gec.d_cumulative, gec.d_base, sizeof (cufftDoubleComplex) * n/2, cudaMemcpyDeviceToDevice));
  memcpy( (void *) gec.h_cumulative_int, (void *) gec.h_base_int, sizeof(int2) * n/2 );
}

/**************************************************************************
 *                                                                        *
 *                        PRP Test Iteration Loop                         *
 *                                                                        *
 **************************************************************************/
float prp_square (int q, int n, int error_flag, int iter = -1)
{
  float terr = 0.0f;

  cufftExecZ2Z (g_plan, (cufftDoubleComplex *) d_x, (cufftDoubleComplex *) d_x, CUFFT_FORWARD);
  square <<< n / 2 / 128, 128 >>> (n, d_x);
  cufftExecZ2Z (g_plan, (cufftDoubleComplex *) d_x, (cufftDoubleComplex *) d_x, CUFFT_INVERSE);
  unweight_and_partial_carry_unsafe_opt <<< n / 2 / 128, 128 >>> (n, d_x, d_negacyclic_weight, d_irrational_weight_inverse, d_xint, d_base_bit, d_block_carry);
  finish_carry_and_reweight<<< n / 2 / 128, 128 >>>(n, d_x, d_negacyclic_weight, d_irrational_weight, d_xint, d_base_bit, d_block_carry);

  //if(error_flag) cudaMemcpyFromSymbol (&terr, g_err, sizeof (float));

  return (terr);
}

/**************************************************************************
 *                                                                        *
 *                        Benchmarking and Testing                        *
 *                                                                        *
 **************************************************************************/
int isReasonable(int fft) // cuFFT has efficient implementations for 7-smooth numbers. Try only those.
{
  int i;

  while(!(fft & 1)) fft >>= 1;
  for(i = 3; i <= 7; i += 2) while((fft % i) == 0) fft /= i;
  return (fft);
}

int isprime(unsigned int n)
{
  unsigned int i;

  if(n<=1) return 0;
  if(n>2 && n%2==0)return 0;

  i=3;
  while(i*i <= n && i < 0x10000)
  {
  if(n%i==0)return 0;
  i+=2;
  }
  return 1;
}

unsigned nextprime(unsigned n)
{
  while( !isprime(n) ) n++;
  return n;
}

void cufftbench (int cufftbench_s, int cufftbench_e, int passes, int device_number)
{

  cudaEvent_t start, stop;
  float outerTime;
  int i, j, k;
  int end = cufftbench_e - cufftbench_s + 1;
  float best_time;
  float *total, *max_diff;

  int n = cufftbench_e << 10;
  int warning = 0;
  cudaError err = cudaSuccess;

  printf ("CUDA bench, testing reasonable fft sizes %dK to %dK, doing %d passes.\n", cufftbench_s, cufftbench_e, passes);
  j = 1;
  while(j < cufftbench_s) j <<= 1;
  k = j;
  while(j < cufftbench_e) j <<= 1;
  if(k > cufftbench_s || j > cufftbench_e) warning = 1;

  total = (float *) malloc (sizeof (float) * end);
  max_diff = (float *) malloc (sizeof (float) * end);

  for(i = 0; i < end; i++)
  {
    total[i] = max_diff[i] = 0.0f;
  }

  cutilSafeCall (cudaEventCreate (&start));
  cutilSafeCall (cudaEventCreateWithFlags (&stop, cudaEventBlockingSync));
  for (j = cufftbench_s; j <= cufftbench_e; j++)
  {
    if(isReasonable(j) <= 1)
    {
      n = j * 1024;
      unsigned q = nextprime(n * 16 + 1);
      int2 *x_int = (int2 *) malloc (sizeof (int2) * n/2);
      alloc_gpu_mem(n);
      write_gpu_data(q, n);
      cufftSafeCall (cufftPlan1d (&g_plan, n / 2, CUFFT_Z2Z, 1));
      init_x(x_int, NULL, q, n, 1);

      for(k = 0; k < passes; k++)
      {
        cutilSafeCall (cudaEventRecord (start, 0));
        for (i = 0; i < 50; i++)
        {
          prp_square(q, n, 0); // with error flag, bench is giving too high values.
        }
        cutilSafeCall (cudaEventRecord (stop, 0));
        err = cutilSafeCall1(cudaEventSynchronize (stop));
        if( cudaSuccess != err)
        {
          printf("Fatal error while benchmarking.");
          exit(2);
        }

        cutilSafeCall (cudaEventElapsedTime (&outerTime, start, stop));
        i = j - cufftbench_s;
        outerTime /= 50.0f;
        total[i] += outerTime;
        if(outerTime > max_diff[i]) max_diff[i] = outerTime;
      }
      if(cudaSuccess == err)
      {
        close_test( x_int );
        printf ("fft size = %dK, ave time = %6.4f msec, max-ave = %0.5f\n", j, total[i] / passes, max_diff[i] - total[i] / passes);
        fflush(NULL);
      }
    }
  }
  cutilSafeCall (cudaEventDestroy (start));
  cutilSafeCall (cudaEventDestroy (stop));

  i = end - 1;
  j = 1;
  while(j < end) j <<= 1;
  j >>= 1;
  k = j - cufftbench_s;
  best_time = total[i] + 1000000.0f;
  while(i >= 0)
  {
    if(total[i] > 0.0f && total[i] < best_time) best_time = total[i];
    else if(i != k) total[i] = 0.0f;
    if(i == k)
    {
      j >>= 1;
      k = j - cufftbench_s;
    }
    i--;
  }

  char fftfile[32];
  FILE *fptr;


  char fftfile_bak[64];

  time_t now;
  struct tm *tm_now = NULL;
  char buffer[192];
  char output[4][32];
  int index = 0;

  now = time(NULL);
  tm_now = localtime(&now);
  strftime(output[0], 32, "%d", tm_now);                             //date
  strftime(output[1], 32, "%H", tm_now);                             //hour
  strftime(output[2], 32, "%M", tm_now);                            //min
  strftime(output[3], 32, "%S", tm_now);                            //sec
  for(i = 0; i < 4; i++) index += sprintf(buffer + index,"%s",output[i]);
  sprintf (fftfile, "%s fft.txt", g_dev.name);
  sprintf (fftfile_bak, "%s fft%s.txt", g_dev.name, buffer);
  (void) unlink (fftfile_bak);
  (void) rename (fftfile, fftfile_bak);
    fptr = fopen_and_lock(fftfile, "w");

  if(!fptr)
  {
    printf ("Cannot open %s.\n",fftfile);
    printf ("Device              %s\n", g_dev.name);
    printf ("Compatibility       %d.%d\n", g_dev.major, g_dev.minor);
    printf ("clockRate (MHz)     %d\n", g_dev.clockRate/1000);
    printf ("memClockRate (MHz)  %d\n", g_dev.memoryClockRate/1000);
    printf("\n  fft    max exp  ms/iter\n");
    for(i = 0; i < end; i++)
    {
      if(total[i] > 0.0f)
      {
         int tl = (int) (exp(0.9782989529 * log ((double) cufftbench_s + i)) * 22379.9259682859 / 1.01);
        if(tl % 2 == 0) tl -= 1;
        while(!isprime(tl)) tl -= 2;
        printf("%5d %10d %8.4f\n", cufftbench_s + i, tl, total[i] / passes);
      }
    }
    fflush(NULL);
  }
  else
  {
    fprintf (fptr, "Device              %s\n", g_dev.name);
    fprintf (fptr, "Compatibility       %d.%d\n", g_dev.major, g_dev.minor);
    fprintf (fptr, "clockRate (MHz)     %d\n", g_dev.clockRate/1000);
    fprintf (fptr, "memClockRate (MHz)  %d\n", g_dev.memoryClockRate/1000);
    fprintf(fptr, "\n  fft    max exp  ms/iter\n");
    for(i = 0; i < end; i++)
    {
      if(total[i] > 0.0f)
      {
        int tl = (int) (exp(0.9784876919 * log ((double) cufftbench_s + i)) * 22366.92473079 / 1.01);
        if(tl % 2 == 0) tl -= 1;
        while(!isprime(tl)) tl -= 2;
        fprintf(fptr, "%5d %10d %8.4f\n", cufftbench_s + i, tl, total[i] / passes);
      }
    }
    unlock_and_fclose(fptr);
    if(warning) printf("\nWARNING, the bounds were not both powers of two, results at either end may not be accurate.\n\n");
    printf("Old %s moved to %s.\n", fftfile, fftfile_bak);
    printf("Optimal fft lengths saved in %s.\n", fftfile);
    fflush(NULL);
  }

  free ((char *) total);
  free ((char *) max_diff);
}

/**************************************************************************
 *                                                                        *
 *                          Keyboard Interaction                          *
 *                                                                        *
 **************************************************************************/
void
SetQuitting (int sig)
{
  g_qu = 1;
  sig==SIGINT ? printf( "\tSIGINT") : (sig==SIGTERM ? printf( "\tSIGTERM") : printf( "\tUnknown signal")) ;
  printf( " caught, writing checkpoint.");
}



/**************************************************************
 *
 *      Main Function
 *
 **************************************************************/

int
check (int q)
{
  const int end = (q + 31) / 32;
  const int width = end + 10;
  const int check_point_size = sizeof (unsigned) * width * 3; 

  int2  *x_int = NULL;
  unsigned *x_packed = NULL;
  int n, j, last = q; // we want q iteration in PRP
  float maxerr = 0.0f, terr = 0.0f;
  timeval time0, time1;
  unsigned long long total_time = 0, diff, diff1, diff2;
  unsigned long long time_adj = 0;
  unsigned long long residue;
  unsigned iter_adj = 0;
  int last_chk = 0;
  int error_flag;
  float error_limit = 0.375;
  signal (SIGTERM, SetQuitting);
  signal (SIGINT, SetQuitting);
  printf ("You may experience a small delay on startup due to Just-in-Time Compilation\n\n");

  x_packed = read_checkpoint(q);
  x_int = init_lucas(x_packed, q, n, j, total_time, time_adj, iter_adj);

  if(!x_int) exit (2);

  if(j == 1)
  {
    printf ("Starting W%d fft length = %dK\n", q, n/1024);
  }
  else 
  {
    printf ("\nContinuing W%d @ iteration %d with fft length %dK, %5.2f%% done\n\n", q, j, n/1024, (float) j*100.0/q);
    last_chk = j;
  }
  fflush (stdout);
  gettimeofday (&time0, NULL);

gec_fail_restart_point:

  for (; j <= last; j++) // Main LL loop
  {
    error_flag = 0;
    if (j % g_ri == 0 || j % g_cpi == 0 || j == last) error_flag = 1;
    else if ((j % 100) == 0) error_flag = 2;

    terr = prp_square (q, n, error_flag, j);
 
    if(terr < 0.0f)  //Nvidia, cufft error. abort. user can try rerunning.
    {
      printf("Fatal error. Please try rerunning.\n\n");
      close_test(x_int); // try cleaning up. may fail. we don't care.
      exit(2);
    }

    if (terr > error_limit) // round off error. abort. user can try rerunning.
    {
      printf ("Round off error at iteration = %d, err = %0.5g > %0.5g, fft = %dK.\n", j,  terr, error_limit, n / 1024);
      close_test(x_int); // try cleaning up. may fail. we don't care.
      exit (2);
    }

    if(terr > maxerr) maxerr= terr;
    if (g_pf && (j % g_po) == 0) // polite flag check
    {
      (void) cutilSafeThreadSync();
    }

    if( (j-gec.last_check_iteration) % gec.L1 == 0) // time for GEC iteration.
    {
      int gec_result = gec_accumulate_and_check(q, n, j);

      switch(gec_result)
      {
        case -1: // routine accumulate
          break;
        case 0:
          printf("GEC failure at iteration %d. Rolling back to %d\n", j, gec.last_check_iteration);
          gec_rollback(q, n);
          // should we force a save point?
          j = gec.last_check_iteration + 1;
          goto gec_fail_restart_point;
        case 1:
          printf("GEC passed at iteration %d\n", j);
          if(q-j > 100) next_gec(q,n); // don't bother planning for another GEC if there is not enough iterations left.
          // should we force a save point? 
          break;
      }
    }

    if (!(error_flag & 1) && !g_qu) continue; // nothing special. go to next iteration

    //checkpoint iteration or quitting or cufft error
    // retrieve and prep the residue
    cutilSafeCall (cudaMemcpy (x_int, d_xint, sizeof (int2) * n/2, cudaMemcpyDeviceToHost));
    standardize_digits(x_int, q, n);
    pack_bits(x_int, x_packed, q, n);
    residue = find_residue(x_packed, q, n);

    gettimeofday (&time1, NULL);
    diff = time1.tv_sec - time0.tv_sec;
    diff1 = 1000000 * diff + time1.tv_usec - time0.tv_usec;
    total_time += diff1;
    diff2 = (unsigned long long) (total_time - time_adj) / 1000000.0 * (last - j) / (j - iter_adj);
    time0.tv_sec = time1.tv_sec;
    time0.tv_usec = time1.tv_usec;

    if(j % g_cpi == 0 || g_qu) // time for writing a checkpoint.
    {
      set_checkpoint_data(x_packed, q, j + 1, total_time, time_adj, iter_adj);
      (x_packed+width)[end + 2] = gec.last_cumulative_iteration;
      (x_packed+width)[end + 3] = gec.accumulate_count;
      (x_packed+width)[end + 4] = gec.last_check_iteration;
      standardize_digits(gec.h_base_int, q, n);

      // GEC check iteration will leave the value already in gec.h_cumulative_int. No need to fetch it again
      // But accumulate step will not.
      if(gec.last_check_iteration < j)
      {
        // get GEC running product
        cutilSafeCall (cudaMemcpy (gec.h_cumulative_int, gec.d_cumulative_int, sizeof (int2) * n/2, cudaMemcpyDeviceToHost));
        standardize_digits(gec.h_cumulative_int, q, n);
      }

      pack_bits(gec.h_cumulative_int, x_packed+width, q, n);
      pack_bits(gec.h_base_int, x_packed+2*width, q, n); // GEC base value

      write_checkpoint(x_packed, q, residue);
    }

    if(g_qu) // user interrupted. time to quit
    {
      printf(" Estimated time spent so far: ");
      print_time_from_seconds((unsigned) (total_time / 1000000), NULL,0);
      printf("\n\n");
      j = last + 1;
      break; // stop the test right here.
    }

    // screen output
    if(g_output_interval)
      process_output(q, n, j, last_chk, maxerr, residue, diff, diff1, diff2, total_time);
    else
    {
      printf ("Iteration %d ", j);
      printbits(residue, q, n, NULL);
      printf (", error = %5.5f, real: ", maxerr);
      print_time_from_seconds ((unsigned) diff, NULL, 0);
      printf (", %4.4f ms/iter, ETA: ", diff1 / 1000.0 / (j - last_chk));
      print_time_from_seconds ((unsigned) diff2, NULL, 0);
      printf (", %5.2f%%\n", (float) j/q*100);
    }
    fflush (stdout);

    last_chk = j;
    reset_err();
    maxerr = 0.0f;
  } /* end main LL for-loop */

  if (!g_qu) // done with test
  { 
    gettimeofday (&time1, NULL);
    FILE* fp = fopen_and_lock(g_RESULTSFILE, "a");
    if(!fp)
    {
      fprintf (stderr, "Cannot write results to %s\n\n", g_RESULTSFILE);
      exit (1);
    }

    /* retrieve and prep the residue */
    cutilSafeCall (cudaMemcpy (x_int, d_xint, sizeof (int) * n, cudaMemcpyDeviceToHost));
    standardize_digits(x_int, q, n);
    pack_bits(x_int, x_packed, q, n);
    residue = find_residue(x_packed, q, n);

    if (residue == 9)
    {
      printf ("W( %d )P, n = %dK, %s", q, n / 1024, program);
      if (fp) fprintf (fp, "W( %d )P, n = %dK, %s", q, n / 1024, program);
    }
    else
      printbits(residue, q, n, fp, x_packed);
    diff = time1.tv_sec - time0.tv_sec;
    total_time +=  diff * 1000000 + time1.tv_usec - time0.tv_usec;
    printf (", estimated total time = ");
    print_time_from_seconds((unsigned) (total_time / 1000000), NULL, 0);

    fprintf(fp, "\n");
    unlock_and_fclose(fp);
    fflush (stdout);
    if(residue != 9) rm_checkpoint (q); // retain check point for PRP, for quick reverification
    g_fft_count = 0;
    printf("\n\n");
  }

  // cleanup
  close_test(x_int);
  free ((char *) x_packed);
  x_packed = NULL;
  return (0);
}


void parse_args(int argc, char *argv[], int* q, int* cufftbench_s, int* cufftbench_e, int* cufftbench_d, int* cufftbench_m);
    /* The rest of the opts are global */

void encode_output_options(void)
{
  int i = 0, j = 0, temp;
  unsigned k;
  char token[196];
  char c;

  c = g_output_string[0];
  while(c)
  {
  if(c == '%')
  {
    i++;
    c = g_output_string[i];
    switch (c)
    {
    case 'a' : //day of week
           g_output_code[j] = 0;
           j++;
           break;
    case 'M' : //month
           g_output_code[j] = 1;
           j++;
           break;
    case 'D' : //date
           g_output_code[j] = 2;
           j++;
           break;
    case 'h' : //hour
           g_output_code[j] = 3;
           j++;
           break;
    case 'm' : //minutes
           g_output_code[j] = 4;
           j++;
           break;
    case 's' : //seconds
           g_output_code[j] = 5;
           j++;
           break;
    case 'y' : //year
           g_output_code[j] = 6;
           j++;
           break;
    case 'z' : //time zone
           g_output_code[j] = 7;
           j++;
           break;
    case 'p' : //testing exponent
           g_output_code[j] = 8;
           j++;
           break;
    case 'P' : //testing Mersenne
           g_output_code[j] = 9;
           j++;
           break;
    case 'f' : //fft as a multiple of K
           g_output_code[j] = 10;
           j++;
           break;
    case 'F' : //fft
           g_output_code[j] = 11;
           j++;
           break;
    case 'i' : //iteration number
           g_output_code[j] = 12;
           j++;
           break;
    case 'o' : //offset
           g_output_code[j] = 13;
           j++;
           break;
    case 'r' : //residue
           g_output_code[j] = 14;
           j++;
           break;
    case 'R' : //residue with upper case hex
           g_output_code[j] = 15;
           j++;
           break;
    case 'C' : //time since last checkpoint hh:mm:ss
           g_output_code[j] = 16;
           j++;
           break;
    case 'T' : //eta d:hh:mm:ss
           g_output_code[j] = 17;
           j++;
           break;
    case 'N' : //residue with upper case hex
           g_output_code[j] = 18;
           j++;
           break;
    case '%' : //residue with upper case hex
           g_output_code[j] = 19;
           j++;
           break;
    case 'x' : //round off error
           g_output_code[j] = 25;
           j++;
           i++;
           temp = g_output_string[i] - 48;
           if(temp > 5 || temp < 0) temp = 5;
           g_output_code[j] = 5 - temp + (temp == 0);
           j++;
           break;
    case 'q' : //iteration timing
           g_output_code[j] = 26;
           j++;
           i++;
           temp = g_output_string[i] - 48;
           if(temp > 5 || temp < 0) temp = 5;
           g_output_code[j] = 5 - temp + (temp == 0);
          j++;
           break;
    case 'c' : //checkpoint interval timing ms/iter
           g_output_code[j] = 27;
           j++;
           i++;
           temp = g_output_string[i] - 48;
           if(temp > 5 || temp < 0) temp = 5;
           g_output_code[j] = 5 - temp + (temp == 0);
           j++;
           break;
    case 'Q' : //checkpoint interval timing iter/sec
           g_output_code[j] = 28;
           j++;
           i++;
           temp = g_output_string[i] - 48;
           if(temp > 5 || temp < 0) temp = 5;
           g_output_code[j] = 5 - temp + (temp == 0);
           j++;
           break;
    case 'd' : //percent completed
           g_output_code[j] = 29;
           j++;
           i++;
           temp = g_output_string[i] - 48;
           if(temp > 5 || temp < 0) temp = 5;
           g_output_code[j] = 5 - temp + (temp == 0);
           j++;
           break;
     default  :
           break;
    }//end switch
    i++;
    c = g_output_string[i];
  }
  else // c != '%', write the string up to the next % to array
  {
    k = 0;
    while(c && c != '%')
    {
    token[k] = c;
    k++;
    i++;
    c = g_output_string[i];
    }
    token[k] = '\0';
    int m = 30, found = 0;
    while(m < 50 && !found)
    {
      if (strlen(g_output[m]) == 0)
      {
        k = (k < 32 ? k : 32);
        strncpy(g_output[m], token, k);
        g_output_code[j] = m;
        j++;
        found = 1;
      }
      else if(!strncmp(token, g_output[m], k) && strlen(g_output[m]) == k)
      {
        g_output_code[j] = m;
        j++;
        found = 1;
      }
      m++;
    }
  }
  }//end while(c)
  g_output_code[j] = -1;
}

int check_interval(int interval)
{
  int k = interval, j = 1;

  while(j <= k) j *= 10;
  j /= 10;
  k =  (int) floor( k / (double) j + 0.5);
  if(k == 3) k = 2;
  else if(k > 3 && k < 8) k = 5;
  else if ( k > 7) k = 10;
  k *= j;
  return k;
}

void process_options(void)
{
#define THREADS_DFLT 256
#define ERROR_ITER_DFLT 100
#define REPORT_ITER_DFLT 10000
#define CHECKPOINT_ITER_DFLT 100000
#define SAVE_FOLDER_DFLT "savefiles"
#define S_F_DFLT 0
#define K_F_DFLT 0
#define D_F_DFLT 0
#define SLEEP_FLAG_DFLT 0
#define SLEEP_VALUE_DFLT 100
#define ROT_F_DFLT 0
#define POLITE_FLAG_DFLT 0
#define POLITE_VALUE_DFLT 1
#define WORKFILE_DFLT "worktodo.txt"
#define RESULTSFILE_DFLT "results.txt"
#define ER_DFLT 85
#define OUTPUT_DFLT "0"
#define HEADER_DFLT "0"
#define HEADERINT_DFLT 15
#define BIG_CARRY_DFLT 0

  char fft_str[192] = "\0";


  if (file_exists(g_INIFILE))
  {
  if( g_df < 0 &&           !IniGetInt(g_INIFILE, "PrintDeviceInfo", &g_df, D_F_DFLT) )
      ;
  if( g_pf < 0 &&            !IniGetInt(g_INIFILE, "Polite", &g_pf, POLITE_FLAG_DFLT) )
    ;
  if( g_po < 0 &&            !IniGetInt(g_INIFILE, "PoliteValue", &g_po, POLITE_VALUE_DFLT) )
    ;
  if( g_dn < 0 &&           !IniGetInt(g_INIFILE, "DeviceNumber", &g_dn, 0) )
    ;
  if( g_cpi < 0 &&          !IniGetInt(g_INIFILE, "CheckpointIterations", &g_cpi, CHECKPOINT_ITER_DFLT) )
      ;
  if( g_ri < 0 &&          !IniGetInt(g_INIFILE, "ReportIterations", &g_ri, REPORT_ITER_DFLT) )
      ;
  if( g_ei < 0 &&          !IniGetInt(g_INIFILE, "ErrorIterations", &g_ei, ERROR_ITER_DFLT) )
      ;
  if( g_fftlen < 0 &&       !IniGetStr(g_INIFILE, "FFTLength", fft_str, "\0") )
    ;
  if( !g_input_file[0] &&    !IniGetStr(g_INIFILE, "WorkFile", g_input_file, WORKFILE_DFLT) )
    ;
  if( !g_RESULTSFILE[0] &&  !IniGetStr(g_INIFILE, "ResultsFile", g_RESULTSFILE, RESULTSFILE_DFLT) )
    fprintf(stderr, "Warning: Couldn't parse ini file option ResultsFile; using default \"%s\"\n", RESULTSFILE_DFLT);
  if( g_output_string[0] < 0 &&  !IniGetStr(g_INIFILE, "OutputString", g_output_string, OUTPUT_DFLT) )
    fprintf(stderr, "Warning: Couldn't parse ini file option OutputString; using default \"%s\"\n", OUTPUT_DFLT);
  if( g_output_header[0] < 0 &&  !IniGetStr(g_INIFILE, "OutputHeader", g_output_header, HEADER_DFLT) )
    fprintf(stderr, "Warning: Couldn't parse ini file option OutputHeader; using default \"%s\"\n", HEADER_DFLT);
  if( g_output_interval < 0 &&  !IniGetInt(g_INIFILE, "OutputHInterval", &g_output_interval, 0) )
    ;
  }
  else fprintf(stderr, "Warning: Couldn't find .ini file. Using defaults for non-specified options.\n");

  //set default values for anything not found in ini file
  if( g_cpi < 0 ) g_cpi = CHECKPOINT_ITER_DFLT;
  if( g_ri < 0 ) g_ri = REPORT_ITER_DFLT;
  if( g_ei < 0 ) g_ei = ERROR_ITER_DFLT;

  if( g_fftlen < 0 ) g_fftlen = 0;
  if( g_dn < 0 ) g_dn = 0;
  if( g_df < 0 ) g_df = D_F_DFLT;
  if( g_pf < 0 ) g_pf = POLITE_FLAG_DFLT;
  if( g_po < 0 ) g_po = POLITE_VALUE_DFLT;
  if( g_output_string[0] < 0) sprintf(g_output_string, OUTPUT_DFLT);
  if( g_folder[0] < 0) sprintf(g_folder, SAVE_FOLDER_DFLT);
  if( !g_input_file[0] ) sprintf(g_input_file, WORKFILE_DFLT);
  if( !g_RESULTSFILE[0] ) sprintf(g_RESULTSFILE, RESULTSFILE_DFLT);
  if( !g_output_header[0] ) sprintf(g_output_header, HEADER_DFLT);
  if( !g_output_string[0] ) sprintf(g_output_string, OUTPUT_DFLT);
  if( g_output_interval < 0 ) g_output_interval = HEADERINT_DFLT;

  if( g_fftlen < 0 ) g_fftlen = fft_from_str(fft_str); // possible if -f not on command line
  int k = check_interval(g_cpi);
  if(k != g_cpi)
  {
    printf("CheckpointIterations = %d from cudaWagstaff.ini must have the form k*10^m for k = 1, 2, or 5.\n",g_cpi);
    printf("Changing to %d.\n", k);
    g_cpi = k;
  }
  k = check_interval(g_ri);
  if(k != g_ri)
  {
    printf("ReportIterations = %d from cudaWagstaff.ini must have the form k*10^m for k = 1, 2, or 5.\n",g_ri);
    printf("Changing to %d.\n", k);
    g_ri = k;
  }
  k = check_interval(g_ei);
  if(k != g_ei)
  {
    printf("ErrorIterations = %d from cudaWagstaff.ini must have the form k*10^m for k = 1, 2, or 5.\n",g_ei);
    printf("Changing to %d.\n", k);
    g_ei = k;
  }
  encode_output_options();
}

int main (int argc, char *argv[])
{
  /* "Production" opts to be read in from command line or ini file */
  int q = -1;
  int f_f;
  int cufftbench_s, cufftbench_e, cufftbench_d, cufftbench_m;

  printf("\n");

  cufftbench_s = cufftbench_e = cufftbench_d = 0;
  cufftbench_m = g_cpi = g_ri = g_fftlen = g_pf = -1;
  g_po = g_df = g_ei = g_dn = -1;
  g_output_string[0] = g_output_header[0] = g_output_interval = g_folder[0] = -1;
  g_input_file[0] = g_RESULTSFILE[0] = 0; /* First character is null terminator */

  parse_args(argc, argv, &q, &cufftbench_s, &cufftbench_e, &cufftbench_d, &cufftbench_m); // The rest of the args are globals
  process_options();
  f_f = g_fftlen; // if the user has given an override... then note this length must be kept between tests
  init_device (g_dn, g_df);
  if (cufftbench_d > 0) cufftbench (cufftbench_s, cufftbench_e, cufftbench_d, g_dn);
  else
  {
  if (q <= 0)
  {
    do
    {
    g_fftlen = f_f; // fftlen and AID change between tests, so be sure to reset them
    if(get_next_assignment(g_input_file, &q, &g_fftlen) || q < 0) exit (2);
    check (q);
    if(!g_qu && clear_assignment(g_input_file, q)) exit (2);
    } while(!g_qu);
  }
  else
  {
    int res = valid_assignment(q, g_fftlen); //! v_a prints warning
    if (!res) exit (2);
    if (res == -1)
    {
    q |= 1;
    while(q > 0 && !isprime(q)) q += 2;
    if(q < 0)
    {
      fprintf(stderr, "Warning: exponent is too big besides!\n");
      exit (2);
    }
    else fprintf(stderr, "Using %d instead.\n", q);
    }
    check (q);
  }
  }
}


void parse_args(int argc,
        char *argv[],
        int* q,
        int* cufftbench_s,
        int* cufftbench_e,
        int* cufftbench_d,
        int* cufftbench_m)
{

  while (argc > 1)
  {
  if (strcmp (argv[1], "-h") == 0) //Help
  {
    fprintf (stderr, "$ cudaWagstaff -h|-v\n\n");
    fprintf (stderr, "$ cudaWagstaff [options] exponent|input_filename\n\n");
    fprintf (stderr, "$ cudaWagstaff [-d device_number] -cufftbench start end passes (see cudaWagstaff.ini)\n\n");
    fprintf (stderr, "               -h          print this help message\n");
    fprintf (stderr, "               -v          print version number\n");
    fprintf (stderr, "               Options\n");
    fprintf (stderr, "               -info       print device information\n");
    fprintf (stderr, "               -d          use the specified device. default 0\n");
    fprintf (stderr, "               -c          checkpoint iterations\n");
    fprintf (stderr, "               -x          screen output iterations\n");
    fprintf (stderr, "               -i          set .ini file name (default = \"cudaWagstaff.ini\")\n");
    fprintf (stderr, "               -f          set fft length (if round off error then exit)\n");
    fprintf (stderr, "               -polite     GPU is polite every n iterations (default -polite 0 = aggressive)\n");
    exit (2);
  }
  else if (strcmp (argv[1], "-v") == 0) //Version
  {
    printf("%s\n\n", program);
    exit (2);
  }
  else if (strcmp (argv[1], "-polite") == 0) // Polite option
    {
    if (argc < 3 || argv[2][0] == '-')
    {
      fprintf (stderr, "can't parse -polite option\n\n");
      exit (2);
    }
    g_po = atoi (argv[2]);
    if (g_po == 0)
    {
      g_pf = 0;
      g_po = 100;
    }
    else g_pf = 1;
    argv += 2;
    argc -= 2;
    }
  else if (strcmp (argv[1], "-d") == 0) // Device number
    {
    if (argc < 3 || argv[2][0] == '-')
    {
      fprintf (stderr, "can't parse -d option\n\n");
      exit (2);
    }
    g_dn = atoi (argv[2]);
    argv += 2;
    argc -= 2;
    }
  else if (strcmp (argv[1], "-i") == 0) //ini file
    {
    if(argc < 3 || argv[2][0] == '-')
    {
      fprintf (stderr, "can't parse -i option\n\n");
      exit (2);
    }
    sprintf (g_INIFILE, "%s", argv[2]);
    argv += 2;
    argc -= 2;
    }
  else if (strcmp (argv[1], "-info") == 0) // Print device info
  {
    g_df = 1;
    argv++;
    argc--;
  }
  else if (strcmp (argv[1], "-cufftbench") == 0) //cufftbench parameters
    {
    if (argc < 5 || argv[2][0] == '-' || argv[3][0] == '-' || argv[4][0] == '-')
    {
      fprintf (stderr, "can't parse -cufftbench option\n\n");
      exit (2);
    }
    *cufftbench_s = atoi (argv[2]);
    *cufftbench_e = atoi (argv[3]);
    *cufftbench_d = atoi (argv[4]);
    argv += 4;
    argc -= 4;
    }
  else if (strcmp (argv[1], "-c") == 0) // checkpoint iteration
    {
    if (argc < 3 || argv[2][0] == '-')
    {
      fprintf (stderr, "can't parse -c option\n\n");
      exit (2);
    }
    g_cpi = atoi (argv[2]);
    if (g_cpi == 0)
    {
      fprintf (stderr, "can't parse -c option\n\n");
      exit (2);
    }
    argv += 2;
    argc -= 2;
    }
  else if (strcmp (argv[1], "-x") == 0) // report iteration
    {
    if (argc < 3 || argv[2][0] == '-')
    {
      fprintf (stderr, "can't parse -x option\n\n");
      exit (2);
    }
    g_ri = atoi (argv[2]);
    if (g_ri == 0)
    {
      fprintf (stderr, "can't parse -x option\n\n");
      exit (2);
    }
    argv += 2;
    argc -= 2;
    }
  else if (strcmp (argv[1], "-f") == 0) //fft length
    {
    if (argc < 3 || argv[2][0] == '-')
    {
      fprintf (stderr, "can't parse -f option\n\n");
      exit (2);
    }
    g_fftlen = fft_from_str(argv[2]);
    argv += 2;
    argc -= 2;
    }
  else // the prime exponent
  {
    if (*q != -1 || strcmp (g_input_file, "") != 0 )
    {
    fprintf (stderr, "can't parse options\n\n");
    exit (2);
    }
    int derp = atoi (argv[1]);
    if (derp == 0)
    {
    sprintf (g_input_file, "%s", argv[1]);
    }
    else *q = derp;
    argv++;
    argc--;
  }
  }
}



